#include "stdafx.h"
#include "dibapi.h"
#include "xword.h"


INLINE DWORD min(DWORD n, DWORD m)		{ return ( (n < m) ? n : m ); }
INLINE DWORD max(DWORD n, DWORD m)		{ return ( (n > m) ? n : m ); }


IMPLEMENT_DYNAMIC(CCrossWordImage, CObject);

#if defined(_DEBUG)
#define new DEBUG_NEW
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

//	BYTE masks for line matching. The reference bit is bit 3.
//
CONST BYTE m0001000 = 0x08;
CONST BYTE m1111000 = 0x78;
CONST BYTE m0001111 = 0x0F;
CONST BYTE m0011000 = 0x18;

//	Other constants
//
CONST BYTE btBLACK = 0x1;
CONST BYTE btWHITE = 0x0;


// Macro for getting debugging options (release build just uses default value)
// TODO: move into commonxx.h
// 
#if _DEBUG_MODE
#define GET_DEBUG_OPTION(name, default)		(GetDebugOption(name, default))
#else
#define GET_DEBUG_OPTION(name, default)		(default)
#endif


//	Local function prototypes
//


//-----------------------------------------------------------------------------
//	Implementation of CCrossWordImage
//-----------------------------------------------------------------------------

CCrossWordImage::CCrossWordImage (VOID)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::CCrossWordImage()\n"));
	m_dwNumRows = 0;
	m_dwNumCols = 0;
	m_pMatrix = NULL;
	m_sFileName = "xword.grid";
}


CCrossWordImage::CCrossWordImage (LPVOID pDIB, LPBYTE pBitmapData, DWORD dwBytesPerRow, DWORD dwNumRows)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::CCrossWordImage(%x, %x, %lu, %lu)\n", pDIB, pBitmapData, dwBytesPerRow, dwNumRows));
	Init(pDIB, pBitmapData, dwBytesPerRow, dwNumRows);
}


VOID CCrossWordImage::Init (LPVOID pDIB, LPBYTE pBitmapData, DWORD dwBytesPerRow, DWORD dwNumRows)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::Init(%x, %x, %lu, %lu)\n", pDIB, pBitmapData, dwBytesPerRow, dwNumRows));
	ASSERT(IS_WIN30_DIB(pDIB));
	LPBITMAPINFOHEADER lpbmi = (LPBITMAPINFOHEADER)pDIB;

	//	Allocate a byte matrix to represent the crossword image
	m_dwNumRows = dwNumRows;
	//m_dwNumCols = 8 * dwBytesPerRow;
	m_dwNumCols = lpbmi->biWidth;
	m_dwMaxDimen = max(m_dwNumRows, m_dwNumCols);
	m_pMatrix = new CByteArray[m_dwNumRows];
	for (DWORD r = 0; r < m_dwNumRows; r++) {
        m_pMatrix[r].SetSize(m_dwNumCols);
		}
	DWORD dwActualBytesPerRow = m_dwNumCols / 8;
	DWORD dwRowPadding = dwBytesPerRow - dwActualBytesPerRow;

	//	Transfer the bit array into a byte array to simplify scanning. This avoids 
	//	tedious bit manipulation (which also can be inefficient). Each byte in the
	//	array will be set to 0x1 if the corresponding bit is dark (usually 0).
	//	NOTE: The bitmap image is stored bottom-up, so we invert it.
	//	TODO: verify that color index 0 equals RGB(0, 0, 0).
	CONST BYTE bitMask[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
	LPBYTE pImageByte = pBitmapData;
    for (INT i = m_dwNumRows - 1; i >= 0; i--) {
        for (INT j = 0; j < dwActualBytesPerRow; j++) {
			BYTE byte = *pImageByte++;
			for (INT k = 0; k < 8; k++) {
				BOOL bDark = ((byte & bitMask[k]) == 0x0);
				m_pMatrix[i].SetAt(j * 8 + k, (BYTE)bDark);
				}
            }
		pImageByte += dwRowPadding;
        }
}


CCrossWordImage::~CCrossWordImage (VOID)
{
	//	Cleanup the mess
	if (m_pMatrix) {
		delete [] m_pMatrix;
		}
}

//	Try to discern the crossword dimensions and structure. This works roughly as follows:
//
//		Transform bit image into byte array (w/ 0x1 for each dark bit)
//		Try to find line end points along each border
//		Compute the likely dimensions of the puzzle (typically 15x15 cells)
//		Compute the size of a grid cell (eg, 24x24 bits if image is 360x360 bits)
//		For each dark bit, increment darkness for corresponding cell
//		For each cell, consider as black if darkness percentage exceeds threshold (eg, 75%)
//
//	NOTES:
//		The main difficulty in analyzing the image is accounting for the likely skew.
//		That is, the lines cannot be assumed to be perfectly straight, especially if
//		a hand scanner was used. Therefore the grid line endpoints are estimated by
//		averaging the coordinates for bits which seem to be on a line (eg, 00X00 or 11X00).
//		This is done for a certain depth along each border (eg, 24 samples of 16 candidate points).
//		(This doesn't work if a row or column is entirely blackened, but so what.)
//
VOID CCrossWordImage::Analyze (VOID)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::Analyze()\n"));
	ASSERT(m_pMatrix);

	//	Dump the bits if desired
	//	TODO: dump the matrix in chunks
	DEBUG_OUT(TL_GARRULOUS + 1, ("The bitmap bits:\n"));
	for (DWORD i = 0; i < m_dwNumRows; i++) {
		CString sRowBits;
		for (DWORD j = 0; j < m_dwNumCols; j++) {
			//DEBUG_OUT(TL_GARRULOUS + 1, ("%c", (m_pMatrix[i][j] + '0')));
			sRowBits += (CHAR)(m_pMatrix[i][j] + '0');
			}
		DEBUG_OUT(TL_GARRULOUS + 1, ("%s\n", (PCSZ)sRowBits));
		}
	DEBUG_OUT(TL_GARRULOUS + 1, ("\n"));

	//	Determine the likely endpoints for the grid lines at each border.
	//	Determine the averages for the min, max and delta
	INT nDepthPercent = GET_DEBUG_OPTION("DepthPercent", 20);
	INT nHorzDepth = MulDiv(m_dwNumCols, nDepthPercent, 100);
	INT nVertDepth = MulDiv(m_dwNumRows, nDepthPercent, 100);
	FindCandidateEndpoints(m_pointsLeft, EDGE_LEFT, nHorzDepth);
	FindCandidateEndpoints(m_pointsTop, EDGE_TOP, nVertDepth);
	FindCandidateEndpoints(m_pointsBottom, EDGE_BOTTOM, nVertDepth);
	FindCandidateEndpoints(m_pointsRight, EDGE_RIGHT, nHorzDepth);

	//	Display the results of the grid lines scanning
	ReconcilePointArrays();

	//	Look for likely blackened squares
	FindCandidateBlackSquares();

	//	Display the final results
	WriteCrosswordDescriptor();
}

inline INT LastX (CPointArray& points)
{
	ASSERT(points.GetSize() > 0);
	INT nValueX = (points.GetSize() > 0) ? points[points.GetUpperBound()].x : -1;
	return (nValueX);
}

inline INT LastY (CPointArray& points)
{
	ASSERT(points.GetSize() > 0);
	INT nValueY = (points.GetSize() > 0) ? points[points.GetUpperBound()].y : -1;
	return (nValueY);
}

VOID CCrossWordImage::ReconcilePointArrays (VOID)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::ReconcilePointArrays()\n"));

	DEBUG_OUT(TL_DETAILED, ("The endpoints for the borders are deemed to be as follows:\n"));
#if _DEBUG_MODE
	//	TODO: define DUMP_POINT_ARRAY
	DumpPointArray(TL_DETAILED, "left: ",  m_pointsLeft, EDGE_LEFT);
	DumpPointArray(TL_DETAILED, "top: ",  m_pointsTop, EDGE_TOP);
	DumpPointArray(TL_DETAILED, "bottom: ",  m_pointsBottom, EDGE_BOTTOM);
	DumpPointArray(TL_DETAILED, "right: ",  m_pointsRight, EDGE_RIGHT);
#endif

	//	Reconcile the points array with each other (ie, add another fudge factor)
	m_nNumHorzPoints = min(m_pointsLeft.GetSize(), m_pointsRight.GetSize());
	m_nNumVertPoints = min(m_pointsTop.GetSize(), m_pointsBottom.GetSize());
	if ((m_nNumHorzPoints == 0) || (m_nNumVertPoints == 0)) {
		DEBUG_OUTF(TL_BASIC, ("Unable to able reconciliation to points arrays\n"));
		return;
		}
#if TRUE
	//	HACK: make sure the points align with each other
	m_pointsLeft.SetSize(m_nNumHorzPoints);
	m_pointsRight.SetSize(m_nNumHorzPoints);
	INT nLeftHorzShift = (m_pointsBottom[0].x - m_pointsTop[0].x);
	INT nRightHorzShift = LastX(m_pointsBottom) - LastX(m_pointsTop);
	for (INT h = 0; h < m_nNumHorzPoints; h++) {
		m_pointsLeft[h].x = m_pointsTop[0].x + MulDiv(h, nLeftHorzShift, m_nNumHorzPoints);
		m_pointsRight[h].x = LastX(m_pointsTop) + MulDiv(h, nRightHorzShift, m_nNumHorzPoints);
		}
	m_pointsTop.SetSize(m_nNumVertPoints);
	m_pointsBottom.SetSize(m_nNumVertPoints);
	INT nTopVertShift = (m_pointsRight[0].y - m_pointsLeft[0].y);
	INT nBottomVertShift = LastY(m_pointsRight) - LastY(m_pointsLeft);
	for (INT v = 0; v < m_nNumVertPoints; v++) {
		m_pointsTop[v].y = m_pointsLeft[0].y + MulDiv(v, nTopVertShift, m_nNumVertPoints);
		m_pointsBottom[v].y = LastY(m_pointsLeft) + MulDiv(v, nBottomVertShift, m_nNumVertPoints);
		}
#endif
	DEBUG_OUT(TL_DETAILED, ("The endpoints after reconciliation:\n"));
#if _DEBUG_MODE
	DumpPointArray(TL_DETAILED, "left: ",  m_pointsLeft);
	DumpPointArray(TL_DETAILED, "top: ",  m_pointsTop);
	DumpPointArray(TL_DETAILED, "bottom: ",  m_pointsBottom);
	DumpPointArray(TL_DETAILED, "right: ",  m_pointsRight);
#endif

	//	Record the delta size
	m_sizeCell = CSize(0, 0);
	if (m_pointsTop.GetSize() > 1) {
		m_sizeCell.cx = m_pointsTop[1].x - m_pointsTop[0].x;
		}
	if (m_pointsLeft.GetSize() > 1) {
		m_sizeCell.cy = m_pointsLeft[1].y - m_pointsLeft[0].y;
		}
	DEBUG_OUT(TL_DETAILED, ("Cell size: %d x %d\n", m_sizeCell.cy, m_sizeCell.cx));
}


//	Return the byte value at the given offset of the grid, accounting for edge type.
//	If the coordinates extend beyond the grid, 0x0 is returned.
//
BYTE CCrossWordImage::GetByte (INT nRefRow, INT nRefCol, EDGE_TYPE eEdgeType, INT nOffset)
{
	DEBUG_OUTF(TL_GARRULOUS + 2, ("CCrossWordImage::GetByte(%d, %d, %d, %d) =>",
			   nRefRow, nRefCol, eEdgeType, nOffset));
	ASSERT(m_pMatrix);

	// Determine the value's coordinates by adding the offset aling the implied dimension
	INT nRow = nRefRow;
	INT nCol = nRefCol;
	if ((eEdgeType == EDGE_LEFT) || (eEdgeType == EDGE_RIGHT)) {
		nRow += nOffset;
		}
	else {
		nCol += nOffset;
		}

	BYTE value = 0x0;					// default to 0
	if (((nRow >= 0) && (nRow < m_dwNumRows))
		&& ((nCol >= 0) && (nCol < m_dwNumCols))) {
		value = m_pMatrix[nRow][nCol];
		}
	DEBUG_OUTF(TL_GARRULOUS + 2, ("0x%02x\n", value));

	return (value);
}


//	Deem whether the bit at the specified location resides on a line, which
//	runs perpendicular to the reference edge. This is guessed by looking at 
//	the context of the bit in the along the specified border.
//
//	The three bits along each side is examined and a line is assumed if
//	any of the following patterns are encountered:
//
//			   b-3 b-2 b-1  b  b+1 b+2 b+3
//				0   0   0   1   0   0   0
//				1   1   1   1   0   0   0
//				0   0   0   1   1   1   1
//				0   0   1   1   0   0   0
//
//	Note that in the case of two consecutive dark bits, only the rightmost
//	bit is assumed to be on a line.
//
//	This doesn't work if the grid size is smaller that 10 bits (but so what).
//
BOOL CCrossWordImage::IsOnLine (INT nRow, INT nCol, EDGE_TYPE eEdge)
{
	ASSERT(m_pMatrix);

	BOOL bIsOnLine = FALSE;
	BOOL bMightOnLine = (m_pMatrix[nRow][nCol] == btBLACK);	// To begin with, the bit must be set
	if (bMightOnLine) {
		BYTE mask = 0x0;
		//	Constuct pattern for bit plus or minus three other bits (eg, 0001000)
		if (0x1 == (GetByte(nRow, nCol, eEdge, -3))) {
			mask |= 0x40;
			}
		if (0x1 == (GetByte(nRow, nCol, eEdge, -2))) {
			mask |= 0x20;
			}
		if (0x1 == (GetByte(nRow, nCol, eEdge, -1))) {
			mask |= 0x10;
			}
		ASSERT(0x1 == (GetByte(nRow, nCol, eEdge, 0)));
		mask |= 0x08;
		if (0x1 == (GetByte(nRow, nCol, eEdge, +1))) {
			mask |= 0x04;
			}
		if (0x1 == (GetByte(nRow, nCol, eEdge, +2))) {
			mask |= 0x02;
			}
		if (0x1 == (GetByte(nRow, nCol, eEdge, +3))) {
			mask |= 0x01;
			}

		//	See if pattern matches a prototypical line pattern
		if ((mask == m0001000)
				|| (mask == m1111000)
				|| (mask == m0001111)
				|| (mask == m0011000)) {
			bIsOnLine = TRUE;
			}
		}

	if (bIsOnLine) {
		DEBUG_OUTF(TL_GARRULOUS + 1, ("CCrossWordImage::IsOnLine(r%d, c%d, e%d) => %d\n",
				   nRow, nCol, eEdge, bIsOnLine));
		}

	return (bIsOnLine);
}


//	Determine the boundaries for the current grid line to be checked. The depth
//	parameter gives the offset from the edge boundary. For instance, for a 360x360
//	byte array to be scanned at the right edge at depth 3, the row bounds are
//	[0, 359] and the columns bounds are [356, 356].
//
VOID CCrossWordImage::DetermineRowColBounds (INT nDepth, EDGE_TYPE eEdge,
											INT& nStartRow, INT& nEndRow, INT& nStartCol, INT& nEndCol)
{
	DEBUG_OUT(TL_GARRULOUS, ("DetermineRowColBounds(%x, %d, %d, _, _, _, _) => ", nDepth, eEdge));
	ASSERT(m_pMatrix);
	//ASSERT(nDepth > 0);

	switch (eEdge) {
		case EDGE_LEFT:
			nStartCol = nDepth;
			nEndCol = nStartCol;
			nStartRow = 0;
			nEndRow = m_dwNumRows - 1;
			break;
		case EDGE_TOP:
			nStartCol = 0;
			nEndCol = m_dwNumCols - 1;
			nStartRow = nDepth;
			nEndRow = nStartRow;
			break;
		case EDGE_RIGHT:
			nStartCol = m_dwNumCols - 1 - nDepth;
			nEndCol = nStartCol;
			nStartRow = 0;
			nEndRow = m_dwNumRows - 1;
			break;
		case EDGE_BOTTOM:
			nStartCol = 0;
			nEndCol = m_dwNumCols - 1;
			nStartRow = m_dwNumRows - 1 - nDepth;
			nEndRow = nStartRow;
			break;
		default:
			ASSERT(0);
			break;
		}
	DEBUG_OUT(TL_GARRULOUS, ("row:[%d, %d] col:[%d, %d]\n", nStartRow, nEndRow, nStartCol, nEndCol));
}


//	Find the mode for a vector of numbers representing counts. The indices are the values in
//	the range [0, m_dwMaxDimen): that is a valid coordinate on the crossword pixel array.
//
INT CCrossWordImage::GetMode (PINT pnCounts)
{
	ASSERT(!IsBadReadPtr(pnCounts, m_dwMaxDimen * sizeof(INT)));
	INT nMode = -1;
	INT nMaxCount = INT_MIN;
	for (INT i = 0; i < m_dwMaxDimen; i++) {
		if (pnCounts[i] > nMaxCount) {
			nMaxCount = pnCounts[i];
			nMode = i;
			}
		}

	return (nMode);
}


//	Given a set of points along a particular border, determine the minimum, maximum,
//	and average values. The points are assumed to be in order. The returned values 
//	are scalars for the coordinate appropriate for the border type (eg, x for top).
//
VOID CCrossWordImage::GetMinMaxAndDelta (CPointArray& endpoints, EDGE_TYPE eEdge, INT& nMin, INT& nMax, INT& nDelta)
{
	DEBUG_OUT(TL_GARRULOUS, ("GetMinMaxAndDelta(%x, %d, %x, %x, %x)\n", endpoints, eEdge, &nMin, &nMax, &nDelta));

	INT nLastValue = 0;
	INT nTotalDelta = 0;
	nMin = 0;
	nMax = 0;
	nDelta = 0;
	PINT anDelta = new INT[m_dwMaxDimen];
	ZeroMemory(anDelta, m_dwMaxDimen * sizeof(INT));
	//	Scan through the list of points, accumulating the deltas (difference from last).
	//	Only the coordinate significant for the border type is used (eg, y for left).
	for (INT i = 0; i < endpoints.GetSize(); i++) {
		INT nValue = 0;
		switch (eEdge) {
#if FALSE
			case EDGE_LEFT:
			case EDGE_RIGHT:
				nValue = endpoints[i].x;
				break;
			case EDGE_TOP:
			case EDGE_BOTTOM:
				nValue = endpoints[i].y;
				break;
#endif
			case EDGE_LEFT:
			case EDGE_RIGHT:
				nValue = endpoints[i].y;
				break;
			case EDGE_TOP:
			case EDGE_BOTTOM:
				nValue = endpoints[i].x;
				break;
			default:
				ASSERT(0);
				break;
			}
		if (i == 0) {
			nMin = nValue;						// determine minimum
			}
		else {
			ASSERT(nValue >= nMin);
			ASSERT(nValue >= nLastValue);
			//nTotalDelta += (nValue - nLastValue);
			INT nDelta = nValue - nLastValue;
			//	Accumulate the counters for each occurrence of the delta
			anDelta[nDelta]++;
			}
		nLastValue = nValue;
		}
	nMax = nLastValue;								// determine maximum
	if (endpoints.GetSize() > 0) {
		//nDelta = (nTotalDelta / endpoints.GetSize());	// determine delta
		nDelta = GetMode(anDelta);
		}
	delete [] anDelta;

	DEBUG_OUT(TL_GARRULOUS, ("GetMinMaxAndDelta: %d %d %d\n", nMin, nMax, nDelta));
}


//	Determine the likely endpoints for the grid lines along a particular border.
//	(The points array is emptied before the points are added.)
//
//	TODO: optimize this to minimize C++ overhead.
//		  Special case each border if necessary to reduce generalization overhead
//
VOID CCrossWordImage::FindCandidateEndpoints (CPointArray& points, EDGE_TYPE eEdge, INT nDepth)
{
	DEBUG_OUT(TL_VERBOSE, ("FindCandidateEndpoints(%x, %d, %d)\n", &points, eEdge, nDepth));
	INT nStartRow = -1;
	INT nEndRow = -1;
	INT nStartCol = -1;
	INT nEndCol = -1;

	//INT nDepth = min(m_dwNumRows, m_dwNumCols) - 1;
	ASSERT(nDepth > 0);
	INT nTotalMin = 0;
	INT nTotalMax = 0;
	INT nTotalDelta = 0;
	INT nNumValid = 0;

	//	Allocate and initialize the counters for determining the mode
	PINT anMin = new INT[m_dwMaxDimen];
	ZeroMemory(anMin, m_dwMaxDimen * sizeof(INT));
	PINT anMax = new INT[m_dwMaxDimen];
	ZeroMemory(anMax, m_dwMaxDimen * sizeof(INT));
	PINT anDelta = new INT[m_dwMaxDimen];
	ZeroMemory(anDelta, m_dwMaxDimen * sizeof(INT));

	//	Check the min/max/delta at each depth in from the border
	for (INT i = 0; i < nDepth; i++) {
		CPointArray ends;
		DetermineRowColBounds(i, eEdge, nStartRow, nEndRow, nStartCol, nEndCol);
		for (INT nRow = nStartRow; nRow <= nEndRow; nRow++) {
			for (INT nCol = nStartCol; nCol <= nEndCol; nCol++) {
				if (IsOnLine(nRow, nCol, eEdge)) {
					//ends.Add(CPoint(nRow, nCol));
					ends.Add(CPoint(nCol, nRow));
					}
				}
			}
#if _DEBUG_MODE
		DumpPointArray(TL_GARRULOUS, "ends: ",  ends, eEdge);
#endif

		//	Compute averages along the given border
		//	TODO: work in units of points instead of scalars (ie, ptMin, ptMax, ptDelta)
		INT nMin = 0;
		INT nMax = 0;
		INT nDelta = 0;
		GetMinMaxAndDelta(ends, eEdge, nMin, nMax, nDelta);
		if (nDelta > 0) {
			//nTotalMin += nMin;
			//nTotalMax += nMax;
			//nTotalDelta += nDelta;
			ASSERT((nMin >= 0) && (nMin < m_dwMaxDimen));
			ASSERT((nMax >= 0) && (nMax < m_dwMaxDimen));
			ASSERT((nDelta >= 0) && (nDelta < m_dwMaxDimen));
			nNumValid++;

			//	Accumulate the counters for each occurrence of the min, max & delta
			anMin[nMin]++;
			anMax[nMax]++;
			anDelta[nDelta]++;
			}
		}

	points.RemoveAll();
	if (nNumValid == 0) {
		DEBUG_OUT(TL_VERBOSE, ("***** No candidate points along edge %d\n", eEdge));
		delete [] anMin;
		delete [] anMax;
		delete [] anDelta;
		return;
		}

	//	Use the mode instead of the average, since the average is skewed too much by
	//	bogus numbers. Crossword puzzles are quite regular, so the mode should be valid.
	//INT nAvgMin = nTotalMin / nNumValid;
	//INT nAvgMax = nTotalMax / nNumValid;
	//INT nAvgDelta = nTotalDelta / nNumValid;
	INT nAvgMin = GetMode(anMin);
	INT nAvgMax = GetMode(anMax);
	INT nAvgDelta = GetMode(anDelta);
	ASSERT(nAvgMax >= nAvgMin);
	DEBUG_OUT(TL_VERBOSE, ("edge %d: AvgMin=%d AvgMax=%d AvgDelta=%d\n",
			  eEdge, nAvgMin, nAvgMax, nAvgDelta));

	//	Fill out the point array using the averages 
	//	This actually uses the modes to avoid averages of averages problems.
	//
	//	NOTE: The insignificant dimension for the border is supplied with default values.
	//		  The values are reconciled later.
	//		  For example, 0 used for y values of left edge points.
	INT nNumPoints = ((nAvgMax - nAvgMin) / nAvgDelta) + 1;
	ASSERT(nNumPoints > 0);
#if TRUE
	nNumPoints = 1 + MulDiv(nAvgMax - nAvgMin, 1, nAvgDelta);
	DEBUG_OUT(TL_VERBOSE, ("nNumPoints=%d\n", nNumPoints));
	if (GET_DEBUG_OPTION("UseFixedSize", TRUE)) {
		nNumPoints = 1 + GET_DEBUG_OPTION("FixedSize", 15);
		}
	// Recalculate the average delta to work around round-off errors
	nAvgDelta = MulDiv(nAvgMax - nAvgMin, 10, (nNumPoints - 1));
	DEBUG_OUT(TL_VERBOSE, ("adjusted average delta=%d (/10)\n", nAvgDelta));
	m_sizeAdjustedCell = CSize(nAvgDelta, nAvgDelta);
#endif
	for (INT p = 0; p < nNumPoints; p++) {
		INT pos = MulDiv(p, nAvgDelta, 10);
		switch (eEdge) {
#if FALSE
			case EDGE_LEFT:
				points.Add(CPoint(nAvgMin + p * nAvgDelta, 0)); break;
			case EDGE_RIGHT:
				points.Add(CPoint(nAvgMin + p * nAvgDelta, m_dwNumCols - 1)); break;
			case EDGE_TOP:
				points.Add(CPoint(0, nAvgMin + p * nAvgDelta)); break;
			case EDGE_BOTTOM:
				points.Add(CPoint(0, nAvgMin + p * nAvgDelta)); break;
#endif
			case EDGE_LEFT:
				points.Add(CPoint(0, nAvgMin + pos)); break;
			case EDGE_RIGHT:
				points.Add(CPoint(m_dwNumCols - 1, nAvgMin + pos)); break;
			case EDGE_TOP:
				points.Add(CPoint(nAvgMin + pos, 0)); break;
			case EDGE_BOTTOM:
				points.Add(CPoint(nAvgMin + pos, m_dwNumRows - 1)); break;
			default:
				ASSERT(0);
				break;
			}
		}

	// Clean up the mess
	delete [] anMin;
	delete [] anMax;
	delete [] anDelta;
}


//	Convert a pixel coordinate into grid cell coordinates.
//	TODO: account for potential skew
//
VOID CCrossWordImage::ConvertPixelToCell (DWORD nPixelRow,  DWORD nPixelCol, DWORD& nGridRow, DWORD& nGridCol)
{
	ASSERT((m_sizeCell.cx > 0) && (m_sizeCell.cy > 0));
	nGridRow = (DWORD)-1;
	nGridCol = (DWORD)-1;
	if ((m_sizeCell.cx == 0) || (m_sizeCell.cy == 0)) {
		return;
		}
#if FALSE
	nGridCol = (nPixelCol - m_pointsTop[0].x) / m_sizeCell.cx;
	nGridRow = (nPixelRow - m_pointsLeft[0].y) / m_sizeCell.cy;
#else
	//nGridCol = MulDiv(nPixelCol - m_pointsLeft[0].x, 10, m_sizeAdjustedCell.cx);
	nGridCol = ((nPixelCol - m_pointsLeft[0].x) * 10) / m_sizeAdjustedCell.cx;
	//nGridRow = MulDiv(nPixelRow - m_pointsTop[0].y, 10, m_sizeAdjustedCell.cy);
	nGridRow = ((nPixelRow - m_pointsTop[0].y) * 10) / m_sizeAdjustedCell.cy;
#endif
	DEBUG_OUT(TL_GARRULOUS + 1, ("CCrossWordImage::ConvertPixelToCell(%lu, %lu, %d, %d)\n",
			  nPixelRow, nPixelCol, nGridRow, nGridCol));
}


//	Scan the pixel image and determine the likely black cells. This works by dividing the
//	image into the cell dimensions previously estimated and counting the number of black
//	pixels for each candidate cell. Those with a given percentage of black pixels will be
//	considered dark.
//
VOID CCrossWordImage::FindCandidateBlackSquares (VOID)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::FindCandidateBlackSquares()\n"));
	DWORD dwGridRow = (DWORD)-1;
	DWORD dwGridCol = (DWORD)-1;

	//	Allocate a matrix to hold the counts, and initialize to 0's
	INT nNumGridRows = m_nNumHorzPoints;		// includes extra row for bit bucket
	INT nNumGridCols = m_nNumVertPoints;		// includes extra col for bit bucket
	if ((nNumGridRows <= 0) || (nNumGridCols <= 0)) {
		DEBUG_OUT(TL_BASIC, ("The crossword grid is degenerate\n"));
		return;
		}
	CIntArray* anCount = new CIntArray[nNumGridRows];
	for (INT r = 0; r < nNumGridRows; r++) {
        anCount[r].SetSize(nNumGridCols);
		for	(INT c = 0; c < nNumGridCols; c++) {
			anCount[r][c] = 0;
			}
		}

	//	Count the occurrences of black squares for each grid cell
	for (INT i = 0; i < m_dwNumRows; i++) {
		for (INT j = 0; j < m_dwNumCols; j++) {
			ASSERT((m_pMatrix[i][j] == btBLACK)	|| (m_pMatrix[i][j] == btWHITE));
			if (m_pMatrix[i][j] == btBLACK)	{
				ConvertPixelToCell(i, j, dwGridRow, dwGridCol);
				if ((dwGridRow < nNumGridRows) && (dwGridCol < nNumGridCols)) {
					anCount[dwGridRow][dwGridCol]++;
					}
				else {
					DEBUG_OUT(TL_VERBOSE, ("Ignoring black pixel at (%d, %d)\n",
										   i, j));
					}
				}
			}
		}

	//	Consider cells 75% dark as black squares
	//	TODO: do some sort of statistical analysis (eg, search for two modes)
	INT nPercentDarkRequired = 75;
#if _DEBUG_MODE
	nPercentDarkRequired = GET_DEBUG_OPTION("PercentDarkRequired", 75);
#endif
	// Initialize array of dark bit counts [0 .. (cell_width x cell_height)]
	CDWordArray adwDarkCount;
	adwDarkCount.SetSize(1 + MulDiv(m_sizeAdjustedCell.cx, m_sizeAdjustedCell.cy, 100));
	PDWORD pData = adwDarkCount.GetData();
	ZeroMemory(pData, adwDarkCount.GetSize() * sizeof(adwDarkCount[0]));

	INT nThreshold = MulDiv(m_sizeCell.cx * m_sizeCell.cy, nPercentDarkRequired, 100);
	m_cellsBlackened.RemoveAll();
	DEBUG_OUT(TL_VERBOSE, ("anCount values (blackness threshold %d of %d):\n", 
			  nThreshold, m_sizeCell.cx * m_sizeCell.cy));
	for (dwGridRow = 0; dwGridRow < nNumGridRows; dwGridRow++) {
		for (dwGridCol = 0; dwGridCol < nNumGridCols; dwGridCol++) {
			DEBUG_OUT(TL_VERBOSE, ("%.4d", anCount[dwGridRow][dwGridCol]));
			if (anCount[dwGridRow][dwGridCol] >= nThreshold) {
				//m_cellsBlackened.Add(CPoint(dwGridRow, dwGridCol));
				m_cellsBlackened.Add(CPoint(dwGridCol, dwGridRow));
				DEBUG_OUT(TL_VERBOSE, ("* "));
				}
			else {
				DEBUG_OUT(TL_VERBOSE, ("  "));
			}
			INT nCountPos = min(adwDarkCount.GetUpperBound(), anCount[dwGridRow][dwGridCol]);
			adwDarkCount[nCountPos]++;
			}
		DEBUG_OUT(TL_VERBOSE, ("\n"));
		}
#if _DEBUG_MODE
	DumpPointArray(TL_DETAILED, "black cells:", m_cellsBlackened);
	// Dumb dark counts 
	// TODO: analyze and revise blackened cells as appropriate
	DEBUG_OUT(TL_VERBOSE, ("\n"));
	for (i = 0; i < adwDarkCount.GetSize(); i++) {
		if (adwDarkCount[i] > 0) {
			DEBUG_OUT(TL_VERBOSE, ("%d:%d ", i, adwDarkCount[i]));
		}
	}
	DEBUG_OUT(TL_VERBOSE, ("\n"));
#endif

	//	TODO: this would be a good place to revise the estimated crossword dimensions
	//		by seeing if any of the bit bucket cells are darkened

	//	Cleanup the mess
	delete [] anCount;
}


//	Write a descriptor for the crossword in a form suitable for Cword.
//
VOID CCrossWordImage::WriteCrosswordDescriptor (VOID)
{
	DEBUG_OUT(TL_VERBOSE, ("CCrossWordImage::WriteCrosswordDescriptor()\n"));
	CONST CHAR szGRID_HEADER[] = "GRID\n";
	CONST CHAR szNUM_PAIR_FMT[] = "%d, %d\n";
	CONST CHAR szNEW_LINE[] = "\n";
	CONST CHAR szOPEN_ERROR_FMT[] = "Unable to open %s for writing";

	CStdioFile sFile;
	BOOL bOK = sFile.Open(m_sFileName, CFile::modeCreate | CFile::modeWrite);
	if (!bOK) {
		CString sError;
		sError.Format(szOPEN_ERROR_FMT, (PCSZ)m_sFileName);
		AfxMessageBox(sError);
		return;
	}
	try {
		sFile.WriteString(szGRID_HEADER);
		CString sLine;
		sLine.Format(szNUM_PAIR_FMT, (m_nNumVertPoints - 1), (m_nNumHorzPoints - 1));
		sFile.WriteString(sLine);
		for (INT i = 0; i < m_cellsBlackened.GetSize(); i++) {
			//sLine.Format(szNUM_PAIR_FMT, m_cellsBlackened[i].y, m_cellsBlackened[i].x);
			sLine.Format(szNUM_PAIR_FMT, m_cellsBlackened[i].x, m_cellsBlackened[i].y);
			sFile.WriteString(sLine);
			}
		sFile.WriteString(szNULL);
		sFile.Close();
	}
	catch (CException* pExc) {
		// TODO: customize the error dialog
		pExc->ReportError();
	}
}


//-----------------------------------------------------------------------------
//	Local functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//	Miscellaneous functions
//-----------------------------------------------------------------------------

//	Dump the given point array to the debugging trace stream. This version
//	ignores the "insignificant" dimesion.
//
VOID DumpPointArray (INT nTraceLevel, PCSZ pszPrefix, CPointArray& points, EDGE_TYPE eEdge)
{
	DEBUG_OUT(nTraceLevel, ("%s", pszPrefix));
	for (INT i = 0; i < points.GetSize(); i++) {
		INT nValue = -1;
		switch (eEdge) {
#if FALSE
			case EDGE_LEFT:
			case EDGE_RIGHT:
				nValue = points[i].x;
				break;
			case EDGE_TOP:
			case EDGE_BOTTOM:
				nValue = points[i].y;
				break;
#endif
			case EDGE_LEFT:
			case EDGE_RIGHT:
				nValue = points[i].y;
				break;
			case EDGE_TOP:
			case EDGE_BOTTOM:
				nValue = points[i].x;
				break;
			default:
				ASSERT_ERROR();
				break;
			}
		DEBUG_OUT(nTraceLevel, ("%d ", nValue));
		}
	DEBUG_OUT(nTraceLevel, ("\n"));
}


//	Dump the given point array to the debugging trace stream.
//
VOID DumpPointArray (INT nTraceLevel, PCSZ pszPrefix, CPointArray& points)
{
	DEBUG_OUT(nTraceLevel, ("%s", pszPrefix));
	for (INT i = 0; i < points.GetSize(); i++) {
		DEBUG_OUT(nTraceLevel, ("(%d, %d) ", points[i].x, points[i].y));
		}
	DEBUG_OUT(nTraceLevel, ("%s\n", pszPrefix));
}


//	Obsolete function for displaying a grid imposed upon a crossword image.
//	This assumes a fixed cell dimension (15x15).
//
VOID ShowCrosswordGrid (CDC* pDC, CRect& rcDest)
{
	DEBUG_OUT(TL_VERBOSE, ("ShowCrosswordGrid(%x)\n", pDC));

	//	Draw a dotted rectange around the bitmap
	CBrush brush;
	brush.CreateStockObject(HOLLOW_BRUSH);
	CPen penBorder(PS_DOT, 1, RGB(192, 0, 0));
	CBrush* pOldBrush = pDC->SelectObject(&brush);
	CPen* pOldPen = pDC->SelectObject(&penBorder);
	InflateRect(rcDest, 1, 1);
	pDC->Rectangle(rcDest);
	pDC->SelectObject(pOldBrush);
	pDC->SelectObject(pOldPen);

    //  Draw the crossword grid lines
	//	TODO: work around rounding errors
	CONST INT nNUM_ROWS = 15;
	CONST INT nNUM_COLS = 15;
	//INT nCellHeight = MulDiv(1, RECTHEIGHT(&rcDest), nNUM_ROWS);
	//INT nCellWidth = MulDiv(1, RECTWIDTH(&rcDest), nNUM_COLS);
	CPen penGrid(PS_DOT, 1, RGB(0, 0, 192));
	pOldPen = pDC->SelectObject(&penGrid);
	for (INT i = 0; i <= nNUM_ROWS; i++) {
		INT nHeight = MulDiv(i, RECTHEIGHT(&rcDest), nNUM_ROWS);
    	pDC->MoveTo(0, nHeight);
    	pDC->LineTo(RECTWIDTH(&rcDest), nHeight);
		for (INT j = 0; j <= nNUM_COLS; j++) {
			INT nWidth = MulDiv(j, RECTWIDTH(&rcDest), nNUM_COLS);
    		pDC->MoveTo(nWidth, 0);
    		pDC->LineTo(nWidth, RECTHEIGHT(&rcDest));
    		}
    	}
	pDC->SelectObject(pOldPen);
}
