// MvInfo.h: interface into MediaView
//

#if !defined(_MV_INFO_H)

// Constants
//
CONST CHAR cMV_ENTRY_SEP = ';';


class CMediaView;


//-----------------------------------------------------------------------------
// CMvInfo
// 
// Class for accessing MediaView. 
//
// OpenTitle: Open a MediaView title
//
// OpenTitle: Close the MediaView title
//
// OpenWordWheel: Open the MediaView wordwheel with the given name
// Wordwheels are used for doing keyword lookup
//
// OpenTopicList: Open the MediaView topic list with the given name
// Topic lists are used for doing full-text search (for cross-references)
//
// UseSgml: Indicate whether the results should be formatted with (pseudo) SGML
//

class CMvInfo : public CObject
{
	DECLARE_DYNCREATE(CMvInfo);
private:
	typedef enum { e_None, e_LookupPhrase, e_LookupReferences, e_OpenTitle, 
				   e_CloseTitle, e_OpenWordWheel, e_OpenTopicList, 
				   e_Exit, m_eUnknown } LOOKUP_ACTION;
public:
	CMvInfo (VOID);
	~CMvInfo (VOID);
	BOOL Init();

	// Basic access functions
	BOOL OpenTitle (PCSZ pszTitle);
	BOOL CloseTitle (VOID);
	BOOL OpenWordWheel (PCSZ pszName);
	BOOL OpenTopicList (PCSZ pszName);
	CString LookupPhrase (PCSZ pszPhrase, BOOL bDefinition = FALSE);
	CString LookupDefinitionPhrase (PCSZ pszPhrase)
				{ return (LookupPhrase(pszPhrase, TRUE)); }
	CString LookupCrossReferences (PCSZ pszPhrase);

	// Multithreaded access functions
	BOOL MtOpenTitle (PCSZ pszTitle);
	BOOL MtCloseTitle (VOID);
	BOOL MtOpenWordWheel (PCSZ pszName);
	BOOL MtOpenTopicList (PCSZ pszName);
	CString MtLookupPhrase (PCSZ pszPhrase, BOOL bDefinition = FALSE);
	CString MtLookupDefinitionPhrase (PCSZ pszPhrase)
				{ return (MtLookupPhrase(pszPhrase, TRUE)); }
	CString MtLookupCrossReferences  (PCSZ pszReferences);

	// Support for multithreading
	VOID LookupWorker (VOID);
	BOOL AcquireLock (VOID);
	BOOL ReleaseLock (VOID);

	// Miscellaneous
	BOOL IsValid (VOID);
public:
	BOOL m_bAddSgml;					// true if SGML-formatting is desired
	BOOL m_bDefinition;					// true if looking up definition text
	CString m_sPhrase;					// phrase to lookup in the MV database
	CString m_sResults;					// results from the lookup
	CString m_sTitle;					// file name of the MediaView title
	CString m_sWordWheel;				// name of word wheel (eg, Dictionary)
	CString m_sTopicList;				// name of topic list (eg, Thesaurus)
	BOOL m_bOK;							// boolean status of last command
	HWND m_hWndParent;					// handle of parent window (for dialog)
private:
	LOOKUP_ACTION m_eLookupAction;		// lookup action to be performed
	CMediaView* m_pMediaView;			// MediaView wrapper class
	CCriticalSection m_csMediaView;		// critical section for synchronization
	CEvent m_eventLookup;				// event for lookup coordination
	CEvent m_eventResults;				// event for signalling results ready
	CWinThread* m_pLookupThread;		// the thread for performing lookups
};

MAKE_CLASS_CAST(CMvInfo);


//-----------------------------------------------------------------------------
// CMvLookup
// 
// Class for doing MediaView lookups (designed particularly for Bookshelf)
// This must be used in conjunction with CMvInfo. These are separated for
// both resource efficiency and for multithreaded concerns.
//
// LookupPhrase: Lookup the topics in the database given a phrase keyword
//
// LookupCrossReferences: Lookup the topics that contain a given phrase
//
// UseSgml: Indicate whether the results should be formatted with (pseudo) SGML
//

class CMvLookup : public CObject
{
	DECLARE_DYNAMIC(CMvLookup);
public:
	CMvLookup (CMvInfo* pMvInfo);
	~CMvLookup (VOID);

	CString LookupPhrase (PCSZ pszPhrase, BOOL bDefinition = FALSE);
	CString LookupDefinitionPhrase (PCSZ pszPhrase)
				{ return (LookupPhrase(pszPhrase, TRUE)); }
	CString LookupCrossReferences (PCSZ pszPhrase);
	CString MtLookupPhrase (PCSZ pszPhrase, BOOL bDefinition = FALSE)
				{ return (LookupPhrase(pszPhrase, bDefinition)); }
	CString MtLookupDefinitionPhrase (PCSZ pszPhrase)
				{ return (LookupPhrase(pszPhrase, TRUE)); }
	CString MtLookupCrossReferences  (PCSZ pszReferences)
				{ return (LookupPhrase(pszReferences)); }
	VOID UseSgml (BOOL bUseSgml = TRUE);
	BOOL IsValid (VOID);

private:
	CMvInfo* m_pMvInfo;					// class for MediaView title info
};

MAKE_CLASS_CAST(CMvLookup);

/////////////////////////////////////////////////////////////////////////////

#endif
