// CLexicon.cpp : implementation file
//

#include "stdafx.h"
#include "common.h"
#include "pcommon.h"
#include "parser.h"
#include "lexicon.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

IMPLEMENT_SERIAL(CDefinition, CObject, 1)
IMPLEMENT_SERIAL(CDefGroup, CObject, 1)
IMPLEMENT_SERIAL(CLexicon, CCmdTarget, 1)

#define new DEBUG_NEW

/////////////////////////////////////////////////////////////////////////////
// CDefinition

//-----------------------------------------------------------------------------
//	Function:	CDefinition constructor
//
//		Create a default definition object.
//
//	Arguments: None
//-----------------------------------------------------------------------------

CDefinition::CDefinition (VOID)
{
}
 
//-----------------------------------------------------------------------------
//	Function:	CDefinition constructor
//
//		Create a definition object, given the phrase, its grammatical category,
//		its definition, and an optional sense number.
//
//	Arguments:
//		phrase		Phrase or word begin defined
//		cat			Category for the phrase (eg, noun)
//		def			Definition of the phrase
//		nSense		sense indicator for the definition (eg, 2)
//-----------------------------------------------------------------------------

CDefinition::CDefinition (CONST CString& phrase, CONST CString& cat, CONST CString& def, INT nSense /* = 1*/)
{
	m_phrase = phrase;
	m_cat = cat;
	m_def = def;
	m_nSense = nSense;
}

//-----------------------------------------------------------------------------
//	Function:	CDefinition destructor
//
//		Cleanup the memory for a definition object
//
//	Arguments: None
//-----------------------------------------------------------------------------

CDefinition::~CDefinition (VOID)
{
}
 
VOID CDefinition::Serialize (CArchive& ar)
{
	CObject::Serialize(ar);
	if (ar.IsStoring()) {
		ar << m_phrase << m_cat << m_def << ToLong(m_nSense);
		}
	else {
		LONG lSense;
		ar >> m_phrase >> m_cat >> m_def >> lSense;
		m_nSense = ToInt(lSense);
		}
}

#if defined(_DEBUG)
VOID CDefinition::AssertValid() const
{
	CObject::AssertValid();
	ASSERT(m_phrase != "");
	ASSERT(m_cat != "");
	ASSERT(m_def != "");
	ASSERT(m_nSense >= 0);
}

VOID CDefinition::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);
	DUMP_VAR(m_phrase);
	DUMP_VAR(m_cat);
	DUMP_VAR(m_def);
	DUMP_VAR(m_nSense);
//	dc << m_phrase << m_cat << m_def << m_nSense;
}
#endif //_DEBUG


/////////////////////////////////////////////////////////////////////////////
// CDefGroup

CDefGroup::CDefGroup ()
{
}

CDefGroup::CDefGroup (CONST CString& head)
{
	m_head = head;
}

CDefGroup::~CDefGroup ()
{
	for (INT i = 0; i < m_entries.GetSize(); i++) {
		CDefinition* pdef = ToCDefinition(m_entries.GetAt(i));
		delete pdef;
		}
}

CDefinition* CDefGroup::LookupPhrase (CONST CString& phrase, CONST CString& cat, INT nSense /*= -1*/)
{
	CDefinition* pdef = NULL;

	TRACEX(4, ("CDefGroup::LookupPhrase for phrase %s (%s:%d) in group %s\n", 
			   (PCSZ)phrase, (PCSZ)cat, nSense, (PCSZ)m_head));	
//// should this function require input in uppercase (for efficiency)? 
	CString sPhrase(phrase);
	sPhrase.MakeUpper();
	
	for (INT i = 0; i < m_entries.GetSize(); i++) {
		CDefinition* pcurrent = ToCDefinition(m_entries.GetAt(i));
		TRACEX(4, ("\tPhrase %d = %s\n", i, (PCSZ)pcurrent->GetPhrase()));
		if ((sPhrase == pcurrent->GetPhrase())
				&& (cat == pcurrent->GetCat())
				&& ((nSense == -1) || (nSense == pcurrent->GetSense()))) {
			pdef = pcurrent;
			break;
			}
		}

	return (pdef);
}

INT CDefGroup::GetNumSenses (CONST CString& phrase, CONST CString& cat)
{
	TRACEX(4, ("CDefGroup::GetNumSenses(%s, %s)\n", (PCSZ)phrase, (PCSZ)cat));
	INT nNumSenses = 0;

	for (INT nSense = 0; ; nSense++) {
		CDefinition* pdef = LookupPhrase(phrase, cat, nSense + 1);
		if (!pdef) {
			nNumSenses = nSense;
			break;
			}
		}
		
	return (nNumSenses);
}

VOID CDefGroup::AddPhrase (CONST CString& phrase, CONST CString& cat, CONST CString& def, INT nSense /* = 1 */)
{
	TRACEX(4, ("CDefGroup::AddPhrase (%s, %s, %s, %d)\n", (PCSZ)phrase, (PCSZ)cat, (PCSZ)def, nSense));	
	ASSERT(phrase.Find(m_head) == 0);
	INT nSenseNum = 1 + GetNumSenses(phrase, cat);

	CDefinition* pdef = new CDefinition(phrase, cat, def, nSenseNum);
	m_entries.Add(pdef);
}

BOOL CDefGroup::PhraseHasCat (CONST CString& phrase, CONST CString& cat)
{
	TRACEX(4, ("CDefGroup::PhraseHasCat (%s, %s)\n", (PCSZ)phrase, (PCSZ)cat));
	BOOL bHasCat = FALSE;
	
	for (INT nSense = 1; ; nSense++) {
		CDefinition* pDef = LookupPhrase(phrase, cat, nSense);
		if (!pDef) {
			break;
			}
		if (pDef->GetCat() == cat) {
			bHasCat = TRUE;
			break;
			}
		}

	return (bHasCat);
}

VOID CDefGroup::Serialize (CArchive& ar)
{
	CObject::Serialize(ar);
	if (ar.IsStoring()) {
		ar << m_head;
		}
	else {
		ar >> m_head;
		}
	m_entries.Serialize(ar);
}

#if defined(_DEBUG)
VOID CDefGroup::AssertValid() const
{
	CObject::AssertValid();
	ASSERT(m_head != "");
	ASSERT_VALID(&m_entries);
}

VOID CDefGroup::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);
	DUMPLN();
	DUMP_VAR(m_head);
	DUMP_VAR(m_entries);
//	dc << "m_head = " << m_head << "\n";
//	dc << "m_entries = " << &m_entries << "\n";
}
#endif //_DEBUG


/////////////////////////////////////////////////////////////////////////////
// CLexicon

CLexicon::CLexicon (VOID)
{
}

CLexicon::CLexicon(CONST CString& sFileName)
{
	BuildFromFile(sFileName);
}

CLexicon::~CLexicon (VOID)
{
	POSITION pos = m_groupMap.GetStartPosition();
	while (pos != NULL) {
		CString key;
		CObject* pObject;
		m_groupMap.GetNextAssoc(pos, key, pObject);
		CDefGroup* pdefGroup = ToCDefGroup(pObject);
		delete pdefGroup;
		}
}

BOOL CLexicon::BuildFromFile (CONST CString& filename)
{
	BOOL bOK = FALSE;

	ASSERT(filename != "");
	TRACEX(4, ("CLexicon::BuildFromFile(%s)\n", (PCSZ)filename));
	TRY {
		CStdioFile file (filename, CFile::modeRead | CFile::typeText);
		CHAR szLine[256];
		INT nLine = 0;
		for (;;) {
			LPSTR lpLine = file.ReadString(szLine, STR_LEN(szLine));
			if (!lpLine) {
				break;
				}
//// what if line > 256
			nLine++;
			TRACEX(5, ("L%d: %s\n", nLine, lpLine));
			CStringx line(lpLine);
			line = line.Trim();
//			line = line.RemoveComment(';');
			line.MakeUpper();
			if ((line != "") && (line[0] != ';')) {
//// make this more efficient
				CStringArray tokens;
				TokenizeText(tokens, line, "\t ");
				if (tokens.GetSize() != 3) {
					CString error = "Invalid lexicon specification at " 
									+ filename + " line " + String(nLine);
					AfxMessageBox(error);
////
					return (FALSE);
					}
				AddPhrase(tokens[0], tokens[1], tokens[2]);
				}
			}
		bOK = TRUE;
		}
	CATCH(CFileException, e) {
#if defined(_DEBUG)
		afxDump << "File exception on '" << filename << "'; " << e;
#endif
		}
	END_CATCH

	if (bOK) {
		BuildGroupMap();
		}

	return (bOK);
}

CDefGroup* CLexicon::FindGroup (CONST CString& head)
{
	TRACEX(4, ("CLexicon::FindGroup(%s)\n", (PCSZ)head));
	CObject* pentry;
//// should this function require input in uppercase (for efficiency)? 
	CString sHead(head);
	sHead.MakeUpper();
	BOOL bOK = m_groupMap.Lookup(sHead, pentry);
	
	return (bOK ? ToCDefGroup(pentry) : NULL);
}

CDefGroup* CLexicon::CreateGroup (CONST CString& head)
{
	TRACEX(4, ("CLexicon::CreateGroup(%s)\n", (PCSZ)head));
	CDefGroup* pdefGroup = new CDefGroup(head);
	m_groupMap.SetAt(head, pdefGroup);
	
	return (pdefGroup);
}

CDefinition* CLexicon::LookupPhrase (CONST CString& phrase, CONST CString& cat, INT nSense /*= -1*/)
{
	TRACEX(4, ("CLexicon::LookupPhrase('%s', %s, %d)\n", (PCSZ)phrase, (PCSZ)cat, nSense));
	CDefinition* pdef = NULL;
//// clean up CStringx
	CStringx phr = phrase;
	CString head = ExtractHead(phr);
	CDefGroup* pdefGroup = FindGroup(head);
	if (pdefGroup) {
		pdef = pdefGroup->LookupPhrase(phrase, cat, nSense);
		}

	return (pdef);
}

BOOL CLexicon::PhraseHasCat (CONST CString& word, CONST CString& cat)
{
//	TRACEX(4 ("PhraseHasCat (%s, %s)\n", word, cat));
	TRACE("PhraseHasCat (%s, %s)\n", (PCSZ)word, (PCSZ)cat);
	CStringx phr = word;
	CString head = ExtractHead(phr);
	CDefGroup* pdefGroup = FindGroup(head);
	BOOL bOK = FALSE;
	
	if (pdefGroup) {
		bOK = pdefGroup->PhraseHasCat(word, cat);
		}

	return (bOK);
}

VOID CLexicon::AddPhrase (CONST CString& phrase, CONST CString& cat, CONST CString& def, INT nSense /* = 1 */)
{
	TRACEX(4, ("CLexicon::AddPhrase (%s, %s, %s, %d)\n", (PCSZ)phrase, (PCSZ)cat, (PCSZ)def, nSense));
//// straighten out CStringx conversions
	CStringx phr = phrase;
	CString head = ExtractHead(phr);
	CDefGroup* pdefGroup = FindGroup(head);
	if (!pdefGroup) {
		pdefGroup = CreateGroup(head);
//// redundant definition group collections
		m_groups.Add(pdefGroup);
		m_groupMap.SetAt(head, pdefGroup);
		}
	pdefGroup->AddPhrase(phrase, cat, def);
}

BEGIN_MESSAGE_MAP(CLexicon, CCmdTarget)
	//{{AFX_MSG_MAP(CLexicon)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BOOL CLexicon::BuildGroupMap (VOID)
{
	TRACEX(4, ("CLexicon::BuildGroupMap\n"));
	BOOL bOK = FALSE;
	ASSERT(m_groups.GetSize() > 0);
	for (INT i = 0; i < m_groups.GetSize(); i++) {
		CDefGroup* pgroup = ToCDefGroup(m_groups[i]);
//		ASSERT(pgroup->IsKindOf(RUNTIME_CLASS(CDefGroup)));
		m_groupMap.SetAt(pgroup->GetHead(), pgroup);
		TRACEX(4, ("Adding association (%s,%p)\n", pgroup->GetHead(), pgroup));
		}
		
	return (bOK);
}

VOID CLexicon::Serialize (CArchive& ar)
{
	CObject::Serialize(ar);
	m_groups.Serialize(ar);
	if (ar.IsStoring()) {
		}
	else {
		BuildGroupMap();
		}
}

#if defined(_DEBUG)
VOID CLexicon::AssertValid() const
{
	CObject::AssertValid();
	ASSERT_VALID(&m_groups);
	ASSERT_VALID(&m_groupMap);
}

//-----------------------------------------------------------------------------
//	Function:	Dump
//
//		Dump the object to the specified dump context (eg, debugging window).
//
//	Arguments:
//		CDumpContext& dc		The dump context for the debugging display
//
//-----------------------------------------------------------------------------

VOID CLexicon::Dump (CDumpContext& dc) const
{
	CObject::Dump(dc);
	DUMPLN();
	DUMPLN_OBJ(m_groups);
	DUMPLN_OBJ(m_groupMap);
//	dc << "m_groups = " << &m_groups << "\n";
//	dc << "m_groupMap = " << &m_groupMap << "\n";
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CLexicon message handlers


/////////////////////////////////////////////////////////////////////////////
// helper functions
