// parseman.cpp : implementation file
//

#include "stdafx.h"
#include "common.h"
#include "pcommon.h"
#include "parser.h"
#include "lexicon.h"
#include "concept.h"
#include "casefram.h"
#include "transiti.h"
#include "graph.h"
#include "parseman.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

inline VOID DumpObject (CObject* pObject, INT nLevel = 3, BOOL bInDepth = TRUE)
{
#if defined(_DEBUG)
	ASSERT(pObject);
	if (nTraceLevel >= nLevel) {
		afxDump.SetDepth(bInDepth);
		pObject->Dump(afxDump);
		}
#endif
}

/////////////////////////////////////////////////////////////////////////////
// CParseManager

IMPLEMENT_SERIAL(CParseManager, CCmdTarget, 1)

CParseManager::CParseManager()
{
	m_pCF = NULL;
	m_pgraph = NULL;
	m_plex = NULL;
}


CParseManager::CParseManager (CONST CString& lexicon, CONST CString& graph)
{
	m_pCF = NULL;
	m_pgraph = new CGraph(graph + ".INI");
	m_plex = new CLexicon(lexicon + ".INI");
}


CParseManager::~CParseManager()
{
	if (m_pCF) {
		delete m_pCF;
		}
	if (m_pgraph) {
		delete m_pgraph;
		}
	if (m_plex) {
		delete m_plex;
		}
}


BOOL CParseManager::StartParse (CONST CString& node, CONST CString* pText = NULL)
{
	TRACEX(TL_VERBOSE, ("CParseManager::StartParse(%s, %s, %d)\n",
					    (PCSZ)node, Q(pText)));
	TRACEX(TL_VERBOSE, ("Testing %s %s %s %s %s\n", Q("1"),  Q("2"), Q("3"), Q(psNULL), Q(pszNULL)));
	BOOL bParseOK = FALSE;

	//	Initialize for the parsing
	ASSERT(pText);
	if (m_pCF) {
		delete m_pCF;
		}
	m_pCF = new CCaseFrame (pText ? *pText : sNull);
	m_pCF->SetLexicon(m_plex);

	// Now do the parsing
	bParseOK = ParseIt(node, pText);

	return (bParseOK);
}


//	ParseIt
//
//	Parse the phrase at the current position, updating the case frame as appropriate:
//	On the first call, the phrase is tokenized and the case frame initialized. In all
//	cases, tokens are repeatedly examined to verify compatibility with the possible
//	transitions of the current graph position (node). When there is a match, the
//	case frame and graph are updated and then this function is recursively invoked.
//
//	Consider: SPOT RUNS
//	A simple case of token/transition compatibility would be the word SPOT and the
//	transition class NOUN. The update in this case consists of advancing both the token
//	position (eg, to RUNS) and the graph node position (eg, to VERB).
//
//	NOTE:
//
//	There is special case handling of PUSH(SUBNET) transitions. This is treated as if
//	there is NIL transition to SUBNET, instead of transitioning to the specified 
//	destination. However, when a (special) POP destination node is reached, the
//	original destination from the PUSH statement is used.
//
BOOL CParseManager::ParseIt (CONST CString& node, CONST CString* pText = NULL)
{
	TRACEX(TL_VERBOSE, ("CParseManager::ParseIt(%s, %s)\n",
					    (PCSZ)node, Q(pText)));
	BOOL bOK = FALSE;

	//	Initialize for the parsing
	//	TODO: improve on the clumsy first time initialization
	m_pCF->IncDepth();

	//	Validate the case frame and graph
	ASSERT_VALID(m_pCF);
	ASSERT_VALID(m_pgraph);
	TRACEX(TL_VERBOSE, ("Parsing position %d/%d of [%s] at depth %d\n", 
			   			m_pCF->GetPos(), m_pCF->GetNumTokens(),
			   			(PCSZ)JoinTokens(m_pCF->GetTokenArray()), m_pCF->GetDepth()));

	//	Keep track of whether case frame is really an original
	BOOL bOriginal = m_pCF->IsOriginal();

	CGraphNode* pnode = m_pgraph->GetNodePtr(node);
	if (!pnode) {
		TRACEX(TL_BASIC, ("Invalid node '%s'\n", (PCSTR)node));
		return (FALSE);
		}

	//	Parse the tokens as follows:
	//	Try each possible transition from the current ATN node; when a token 
	//	is acceptable, revise the case frame & graph and recursively parse,
	//	stopping if successful. Otherwise, restore case frame and try again.
	for (INT i = 0; i < pnode->GetNumTransitions(); i++) {

		//	Special case handing of PUSH transition
		//	TODO: make this more straightforward
		CString sPrevPopDestination;
		if (pnode->IsPushLink(i)) {
			CString sPseudoTransLabel = ResolvePtr(pnode->GetTransLabel(i));
			TRACEX(TL_DETAILED, ("Simulating a PUSH operation to %s\n", (PCSZ)sPseudoTransLabel));
			sPrevPopDestination = m_pCF->GetPopDestination();
			CString sToNode = ResolvePtr(pnode->GetToNode(i));
			m_pCF->SetPopDestination(sToNode);
			pnode = m_pgraph->GetNodePtr(sPseudoTransLabel);
			if (!pnode) {
				TRACEX(TL_BASIC, ("Invalid node '%s' in PUSH transition\n", (PCSZ)sPseudoTransLabel));
				return (FALSE);
				}
			}

		//	Get the current transition for the node
		CTransition* ptrans = pnode->GetTransition(i);
		if (!ptrans) {
			// TODO: change String to ToString?; also, add CStringx::FormatString
			CString sError ("Internal error: unable to resolve transition ");
			sError += String(i) + cSPACE + "of " + pnode->GetLabel();
			AfxMessageBox(sError);
			}
		TRACEX(3, ("Trying transition %s (%d/%d of %s) at depth %d\n", 
					(PCSZ)ptrans->GetType(), i + 1, pnode->GetNumTransitions(),
					(PCSZ)pnode->GetLabel(), m_pCF->GetDepth()));
		if (ptrans->CanProceed(m_pCF)) {

			//	Save the current state to allow for backtracking
			//	TODO: optimize the backing up of the frame
			CCaseFrame backupFrame (*m_pCF);

			//	Process the transition and proceed with the parse
			ptrans->UpdateState(m_pCF);
			DumpObject(m_pCF);

			//	Recursively parse the phrase unless a terminal node is reached
			CString* plabel = pnode->GetToNode(i);
			ASSERT(plabel);
			if (plabel) {
				//	Special case check for POP transition
				CString sPrevPushDest;
				if (*plabel == "POP") {
					//	Resume the parse at the destination from the previous PUSH
					//	TODO: implement m_pCF's pop destination as a stack
					sPrevPushDest = m_pCF->GetPopDestination();
					TRACEX(TL_DETAILED, ("Simulating a POP operation to %s\n", (PCSZ)sPrevPushDest));
					m_pCF->SetPopDestination(sPrevPopDestination);
					plabel = &sPrevPushDest;
					}				
					
				//	TODO: change this to plabel->IsTerminal()
				if (*plabel == "STOP") {
					bOK = TRUE;
					}
				else {
					bOK = ParseIt(*plabel);
					}
				}

			//	If the parse has succeeded, stop searching (exit loop)
			//	TODO: rework this loop exit
			if (bOK && (m_pCF->GetNumTokensLeft() > 0)) {
				TRACEX(TL_VERBOSE, ("Failing at depth %d due to excess tokens (%d)\n",
									m_pCF->GetDepth(), m_pCF->GetNumTokensLeft()));
				bOK = FALSE;
				}
			if (bOK) {
				break;
				}
					
			//	Otherwise, restore the saved state
			*m_pCF = backupFrame;
			}
		}

	//	Restore the 'original case frame' status
	m_pCF->SetOriginal(bOriginal);

	return (bOK);
}


VOID CParseManager::Serialize (CArchive& ar)
{
	CObject::Serialize(ar);
	if (ar.IsStoring()) {
		ar << m_pCF << m_plex << m_pgraph;
		}                                
	else {
		ar >> m_pCF >> m_plex >> m_pgraph;
		}
}

#if defined(_DEBUG)
VOID CParseManager::AssertValid() const
{
	CObject::AssertValid();
	if (m_pCF) {
		ASSERT_VALID(m_pCF);
		}
	if (m_plex) {
		ASSERT_VALID(m_plex);
		}
	if (m_pgraph) {
		ASSERT_VALID(m_pgraph);
		}
}

VOID CParseManager::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);
	dc << "\n";
	dc << "m_pCF = " << m_pCF << "\n";
	dc << "m_plex =" << m_plex << "\n";
	dc << "m_pgraph =" << m_pgraph << "\n";
}
#endif //_DEBUG

BEGIN_MESSAGE_MAP(CParseManager, CCmdTarget)
	//{{AFX_MSG_MAP(CParseManager)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CParseManager message handlers

