// transiti.cpp : implementation file
//

#include "stdafx.h"
#include "common.h"
#include "pcommon.h"
#include "parser.h"
#include "concept.h"
#include "lexicon.h"
#include "casefram.h"
#include "transiti.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

IMPLEMENT_SERIAL(CTransition, CObject, 1)
IMPLEMENT_SERIAL(CNilTransition, CTransition, 1)
IMPLEMENT_SERIAL(CCatTransition, CTransition, 1)
IMPLEMENT_SERIAL(CAdjTransition, CCatTransition, 1)
IMPLEMENT_SERIAL(CAdvTransition, CCatTransition, 1)
IMPLEMENT_SERIAL(CDetTransition, CCatTransition, 1)
IMPLEMENT_SERIAL(CNounTransition, CCatTransition, 1)
IMPLEMENT_SERIAL(CVerbTransition, CCatTransition, 1)

#define new DEBUG_NEW

/////////////////////////////////////////////////////////////////////////////
// CTransition

//// should the class registration be changed into object registration

CMapStringToOb CTransition::m_transMap;

CTransition::CTransition (VOID)
{
}

CTransition::CTransition (CONST CString& type)
{
	m_type = type;
	Register(type, this);
}

CTransition::~CTransition (VOID)
{
	Deregister(m_type);
}

//	Register - Register the CTransition-derived class to handle	the specified ATN transition
//

VOID CTransition::Register(CONST CString& name, CTransition* ptrans)
{
	TRACE("CTransition::Register(%s, %p)\n", name, ptrans);
	m_transMap.SetAt(name, ptrans);
}

VOID CTransition::Deregister (CONST CString& name)
{
	m_transMap.RemoveKey(name);
}

//	FindTransition - Find the class that handles the transition
//		assumes class hasn't been deleted (eg, isn't out of scope)
//
CTransition* CTransition::FindTransition (CONST CString& name)
{
	TRACEX(TL_VERBOSE, ("CTransition::FindTransition (%s)\n", (PCSZ)name));
	CObject* pentry;
	BOOL bOK = m_transMap.Lookup(name, pentry);
	
	return (bOK ? ToCTransition(pentry) : NULL);
}

VOID CTransition::Serialize (CArchive& ar)
{
	CObject::Serialize(ar);
	if (ar.IsStoring()) {
		ar << m_type;
		}
	else {
		ar >> m_type;
		}
}

#if defined(_DEBUG)
VOID CTransition::AssertValid() const
{
	CObject::AssertValid();
	ASSERT(m_type != "");
}

VOID CTransition::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);
	dc << "Type = " << m_type << "\n";
}
#endif //_DEBUG


/////////////////////////////////////////////////////////////////////////////
// CNilTransition - always succeeds

CNilTransition::CNilTransition (VOID)
	: CTransition("NIL")
{
}

VOID CNilTransition::Serialize (CArchive& ar)
{
	CTransition::Serialize(ar);
}

#if defined(_DEBUG)
VOID CNilTransition::AssertValid() const
{
	CTransition::AssertValid();
}

VOID CNilTransition::Dump(CDumpContext& dc) const
{
	CTransition::Dump(dc);
}
#endif

/////////////////////////////////////////////////////////////////////////////
// CCatTransition - simple grammatical category (eg, noun)

CCatTransition::CCatTransition (VOID)
	: CTransition ("CAT")
{
}

CCatTransition::CCatTransition (CONST CString& type, CONST CString& cat)
	: CTransition (type)
{
	SetCategory(cat);
}

BOOL CCatTransition::CanProceed (CCaseFrame* pCF)
{
	ASSERT(pCF);
	BOOL bOK = FALSE;

	ASSERT(pCF && pCF->GetLexicon());
	if (pCF) {
		CLexicon* plex = pCF->GetLexicon();
		if (plex && (pCF->GetNumTokensLeft() > 0)) {
			bOK = plex->PhraseHasCat(pCF->GetCurrentToken(), m_cat);
			}
#if FALSE
//// for efficiency: this requires saving state information in the 
////	case frame for the benefit of UpdateState
//// alternatives:	- combine CanProceed & UpdateState
////				- define a transition hint (eg, num tokens = 3)
		if (!bOK) {
			INT nMaxPhraseSize = min(GetMaxPhraseSize(pCF->GetCurrentToken()),
								 pCF->NumTokensLeft());
			for (INT nSize = 2; nSize <= nMaxPhraseSize; nSize++) {
				CString phrase = pCF->GetCurrentTokens(nSize);
				}
			}
#endif
		}

	TRACEX(TL_VERBOSE, ("%d <= CCatTransition::CanProceed(%p) for %s\n", bOK, pCF, (PCSZ)m_cat));	

	return (bOK);
}

VOID CCatTransition::UpdateState (CCaseFrame* pCF)
{
	TRACEX(TL_VERBOSE, ("CCatTransition::UpdateState(%p)\n", pCF));
	ASSERT(pCF);
	pCF->AdvanceToken();
#if FALSE
	if (pHint->nTokenLen > 1) {
		pCF->AdvanceToken(pHint->nTokenLen - 1);
		}
#endif	
}

VOID CCatTransition::Serialize (CArchive& ar)
{
	CTransition::Serialize(ar);
	if (ar.IsStoring()) {
		ar << m_cat;
		}
	else {
		ar >> m_cat;
		}
}

#if defined(_DEBUG)
VOID CCatTransition::AssertValid() const
{
	CTransition::AssertValid();
}

VOID CCatTransition::Dump(CDumpContext& dc) const
{
	CTransition::Dump(dc);
	dc << "Cat = " << m_cat;
}
#endif

//------------------------------------------------------------------------------

CAdjTransition::CAdjTransition (VOID)
	: CCatTransition ("ADJ", "ADJ")
{
}

#if FALSE
BOOL CAdjTransition::CanProceed (CCaseFrame* pCF)
{
	return (CCatTransition::CanProceed(pCF));
}

VOID CAdjTransition::UpdateState (CCaseFrame* pCF)
{
	CCatTransition::UpdateState(pCF);
}

VOID CAdjTransition::Serialize (CArchive& ar)
{
	CCatTransition::Serialize(ar);
}
#endif

#if defined(_DEBUG)
VOID CAdjTransition::AssertValid() const
{
	CCatTransition::AssertValid();
}

VOID CAdjTransition::Dump(CDumpContext& dc) const
{
	CCatTransition::Dump(dc);
}
#endif

//------------------------------------------------------------------------------

CAdvTransition::CAdvTransition (VOID)
	: CCatTransition ("ADV", "ADV")
{
}

#if FALSE
BOOL CAdvTransition::CanProceed (CCaseFrame* pCF)
{
	return (CCatTransition::CanProceed(pCF));
}

VOID CAdvTransition::UpdateState (CCaseFrame* pCF)
{
	CCatTransition::UpdateState(pCF);
}

VOID CAdvTransition::Serialize (CArchive& ar)
{
	CCatTransition::Serialize(ar);
}
#endif

#if defined(_DEBUG)
VOID CAdvTransition::AssertValid() const
{
	CCatTransition::AssertValid();
}

VOID CAdvTransition::Dump(CDumpContext& dc) const
{
	CCatTransition::Dump(dc);
}
#endif

//------------------------------------------------------------------------------

CDetTransition::CDetTransition (VOID)
	: CCatTransition ("DET", "DET")
{
}

#if FALSE
BOOL CDetTransition::CanProceed (CCaseFrame* pCF)
{
	return (CCatTransition::CanProceed(pCF));
}

VOID CDetTransition::UpdateState (CCaseFrame* pCF)
{
	CCatTransition::UpdateState(pCF);
}

VOID CDetTransition::Serialize (CArchive& ar)
{
	CCatTransition::Serialize(ar);
}
#endif

#if defined(_DEBUG)
VOID CDetTransition::AssertValid() const
{
	CCatTransition::AssertValid();
}

VOID CDetTransition::Dump(CDumpContext& dc) const
{
	CCatTransition::Dump(dc);
}
#endif

//------------------------------------------------------------------------------


CNounTransition::CNounTransition (VOID)
	: CCatTransition ("NOUN", "NOUN")
{
}

#if FALSE
BOOL CNounTransition::CanProceed (CCaseFrame* pCF)
{
	return (CCatTransition::CanProceed(pCF));
}

VOID CNounTransition::UpdateState (CCaseFrame* pCF)
{
	CCatTransition::UpdateState(pCF);
}

VOID CNounTransition::Serialize (CArchive& ar)
{
	CCatTransition::Serialize(ar);
}
#endif

#if defined(_DEBUG)
VOID CNounTransition::AssertValid() const
{
	CCatTransition::AssertValid();
}

VOID CNounTransition::Dump(CDumpContext& dc) const
{
	CCatTransition::Dump(dc);
}
#endif

//------------------------------------------------------------------------------

CVerbTransition::CVerbTransition (VOID)
	: CCatTransition ("VERB", "VERB")
{
}

#if FALSE
BOOL CVerbTransition::CanProceed (CCaseFrame* pCF)
{
	return (CCatTransition::CanProceed(pCF));
}

VOID CVerbTransition::UpdateState (CCaseFrame* pCF)
{
	CCatTransition::UpdateState(pCF);
}

VOID CVerbTransition::Serialize (CArchive& ar)
{
	CCatTransition::Serialize(ar);
}
#endif

#if defined(_DEBUG)
VOID CVerbTransition::AssertValid() const
{
	CCatTransition::AssertValid();
}

VOID CVerbTransition::Dump(CDumpContext& dc) const
{
	CCatTransition::Dump(dc);
}
#endif

//------------------------------------------------------------------------------

//	Create static instances of each transition
//
CNilTransition trNil;
CCatTransition trCat;
CAdjTransition trAdj;
CAdvTransition trAdv;
CDetTransition trDet;
CNounTransition trNoun;
CVerbTransition trVerb;
