// ReadDict.cpp
//
// Test application for reading TRL dictionary (in ascii) and verifying the
// Lookup module (Lookup.dll)
//

#include "common.h"
#include "lookup.h"

CONST CHAR cSLASH = 		'/';
CONST CHAR cSTART_PHRASE = 0x02;
CONST CHAR cEND_PHRASE = 0x06;
CONST CHAR szDEF_DICT[] = "E:\\DATA\\DICT.DAT";
CONST CHAR cDEF_FIRST_LETTER = 'A';
CONST CHAR cDEF_LAST_LETTER = 'A';
CONST CHAR szOUTPUT_FILE_FMT[] = "DICT_%c.OUT";
CONST INT nMAX_LINE_LEN = 255;
CONST INT nPHRASE_OUTLEN = 25;
CONST LONG lMAX_PHRASES = 1000;

// Globals
//
CWinApp the_app;


inline LONG ToLong (PCSZ pszText)
{
	return (StringToLong(pszText));
}


CString FilterGarbage (PCSZ pszText)
{
	CString sText;

	for (CHAR cPrev = cNULL; *pszText; pszText++) {
		CHAR cNext = *pszText;
		if (!MyIsPrintable(cNext)) {
			continue;
			}
		if (IsSpace(cNext)) {
			cNext = cSPACE;			// make it a space proper
			if (cNext == cPrev) {
				continue;			// ignore successive blanks
				}
			}
		sText += cNext;
		cPrev = cNext;
    	}

	return (sText);
}



VOID WriteEntry (CStdioFile& output, CString& sPhrase, CString& sEntry, BOOL bUseSpaces, INT nEntryMax)
{
	output.WriteString(sPhrase);
	if (bUseSpaces) {
		INT nBlankLen = max(nPHRASE_OUTLEN - sPhrase.GetLength(), 1);
		CString sBlank (cSPACE, nBlankLen);
		output.WriteString(sBlank);
		DEBUG_OUT(TL_VERBOSE, ("%s%s", (PCSZ)sPhrase, (PCSZ)sBlank));
		}
	else {	
		output.WriteString(szTAB);
		DEBUG_OUT(TL_VERBOSE, ("%s\t", (PCSZ)sPhrase));
		}
		
	//	Output the entry text
	CString sOut = FilterGarbage(sEntry);
	if (sOut.GetLength() > nEntryMax) {
		sOut = sOut.Left(nEntryMax - 3) + "...";
		}
	output.WriteString(sOut);
	output.WriteString(szNEW_LINE);
	DEBUG_OUT(TL_GARRULOUS, ("%.80s", (PCSZ)sOut));
	DEBUG_OUT(TL_VERBOSE, ("\n"));
}


INT main (INT nArgc, PSTR apszArgv[])
{
	INIT_DEBUGGING();
	//INIT_COMMON(GetModuleHandle(NULL), NULL, "ReadDict");
	AfxWinInit((HINSTANCE)0x400000, NULL, "", SW_SHOW);
	the_app.InitApplication();
	the_app.InitInstance();
	INIT_COMMON(the_app.m_hInstance, NULL, "ReadDict");

	LONG lMaxEntries = lMAX_PHRASES;
	CHAR cFirstLetter = cDEF_FIRST_LETTER;
	CHAR cLastLetter = cDEF_LAST_LETTER;
	CString sDict (szDEF_DICT);
	BOOL bUseSpaces = TRUE;
	INT nEntryMax = nMAX_LINE_LEN;
	BOOL bUseFile = FALSE;				// use output file (DICT_[A-Z].OUT)
	BOOL bVerify = FALSE;				// verify Lookup on entries

	// Initialize the output files
	CStdioFile dict;
	//CStdioFile output (GetStdHandle(STD_OUTPUT_HANDLE));
	CStdioFile error (stderr);
	CStdioFile output (stdout);

	// Do command line processing
	// TODO: massive rework of command-line parsing (fix trailing dash bug)
	ParseCmdLine(apszArgv, nArgc);		// common arguments (eg, trace level)
	for (INT i = 1; i < nArgc; i++) {
		PCSZ pszArg = apszArgv[i];
		if (IsCommandSwitchDelim(pszArg[0])) {
			switch (ToUpper(pszArg[1])) {
				case 'B': {
					cFirstLetter = pszArg[2];
					DEBUG_OUT(TL_VERBOSE, ("First letter (/B): %c\n", cFirstLetter));
					break;
					}
				case 'E': {
					cLastLetter = pszArg[2];
					DEBUG_OUT(TL_VERBOSE, ("Last letter (/E): %c\n", cLastLetter));
					break;
					}
				case 'L': {
					nEntryMax = (INT)ToLong(pszArg + 2);
					DEBUG_OUT(TL_VERBOSE, ("Max entry length (/L): %d\n", nEntryMax));
					break;
					}
				case 'M': {
					lMaxEntries = ToLong(pszArg + 2);
					DEBUG_OUT(TL_VERBOSE, ("Max entries (/M): %ld\n", lMaxEntries));
					break;
					}
				case 'S': {
					bUseSpaces = (pszArg[2] != cDASH);
					DEBUG_OUT(TL_VERBOSE, ("Use spaces (/S): %d\n", bUseSpaces));
					break;
					}
				case 'D': {
					sDict = pszArg + 2;
					DEBUG_OUT(TL_VERBOSE, ("Dictionary (/D): %s\n", (PCSZ)sDict));
					break;
					}
				case 'F': {
					bUseFile = TRUE;
					break;
					}
				case 'V': {
					bVerify = TRUE;
					break;
					}
				case '?':
					error.WriteString("Format: readdict [/B<start char>] [/E<end char>] [/L<entry length limit>] [/M<max entries>] [/S] [/D<dict>]\n");
					return (1);
				default:
					break;
				}
			}
		}
	if (cFirstLetter > cLastLetter) {
		cLastLetter = cFirstLetter;
		}
	DEBUG_OUT(TL_DETAILED, ("Reading dictionary '%s' from %c to %c (limit=%d)\n",
			  (PCSZ)sDict, cFirstLetter, cLastLetter, lMaxEntries));

	// Open the dictionary through Lookup if verification is desired
	CLookup* pLookup = NULL;
	if (bVerify) {
		pLookup = new CLookup();
		// TODO: add bMultithreded = FALSE override to InitLookup
		pLookup->InitSimpleLookup(TRUE, FALSE, FALSE); // TRL; WordNet; BookShelf
		}

	// Open the dictionary
	// TODO: figure out the stupid sharing problem
	BOOL bOK = dict.Open(sDict, CFile::modeRead | CFile::typeText | CFile::shareDenyNone);
	if (!bOK) {
		CString sError;
		sError.Format("Unable to open dictionary (%s)\n", (PCSZ)sDict);
		error.WriteString(sError);
		return (1);
		}

	//	Read the entries
	BOOL bNew = TRUE;
	CString sPhrase;
	CString sEntry;
	LONG lCount = 0;
	BOOL bNewLetter = TRUE;
	CHAR cCurrentLetter = cSPACE;

	for (;;) {
		//	Read the next line
		CString sLine;
		PSZ pszLine = sLine.GetBuffer(nMAX_LINE_LEN);
		dict.ReadString(pszLine, nMAX_LINE_LEN);
		sLine.ReleaseBuffer();
		if (sLine.GetLength() == 0) {
			DEBUG_OUT(TL_BASIC, ("End of file encountered after %d phrases\n", lCount));
			break;
			}
		DEBUG_OUT(TL_GARRULOUS + 1, ("%.128s\n", (PCSZ)FilterText(sLine)));
		bNew = (sLine[0] == cSTART_PHRASE);

		//	Create a new one entry if start character found
		if (bNew) {

			//	Output the current entry if any
			if (!sEntry.IsEmpty()) {
				//	Create the results file when a new letter is encountered
				if (bNewLetter) {
					if (bUseFile) {
						if (output.m_hFile != CFile::hFileNull) {
							output.Close();
							}
						CHAR szOutFile[MAX_PATH];
						sprintf(szOutFile, szOUTPUT_FILE_FMT, cCurrentLetter);
						DEBUG_OUT(TL_DETAILED, ("Opening new output file '%s'\n", szOutFile));
						// TODO: figure out the stupid sharing problem
						output.Open(szOutFile, CFile::modeCreate | CFile::modeWrite | 
											   CFile::typeText | CFile::shareDenyNone);
						}
					bNewLetter = FALSE;
					}

				//	Output the phrase
				if (cCurrentLetter >= cFirstLetter) {
					lCount++;
					ASSERT(output.m_hFile != CFile::hFileNull);
					WriteEntry(output, sPhrase, sEntry, bUseSpaces, nEntryMax);

					// (optional) Verify that Lookup returns the same entry
					if (bVerify) {
						sPhrase.TrimRight();
						CString sDefinition = pLookup->GetDefinitionText(sPhrase);
						if (!sDefinition.IsEmpty()) {
							// TODO: verify that the entry is the same
							CString sDummyHead ("->");
							if (Pos(sDefinition, sPhrase) == 0) {
								sDefinition = sDefinition.Mid(sPhrase.GetLength());
								}
							WriteEntry(output, sDummyHead, sDefinition, FALSE, nEntryMax);
							}
						else {
							CString sError;
							sError.Format("*** Unable to retrieve definition for %s\n", 
										  (PCSZ)sPhrase);
							output.WriteString(sError);
							}
						}
					}
				else {
					DEBUG_OUT(TL_VERBOSE, ("Ignoring %s\n", (PCSZ)sPhrase));
					}
				sEntry.Empty();
				}
				
			//	Get the next phrase
			ASSERT(sLine[0] == cSTART_PHRASE);
			INT nStartPhrase = 1;
			SkipSpaces(sLine, nStartPhrase);
			INT nEndPhrase = sLine.Find(cEND_PHRASE);
			ASSERT(nEndPhrase != nINVALID_INDEX);
			INT nPhraseLen = nEndPhrase - nStartPhrase;
			sPhrase = FilterGarbage(sLine.Mid(nStartPhrase, nPhraseLen));
			//ASSERT(sPhrase.GetLength() > 0);
			if (sPhrase.IsEmpty()) {
				DEBUG_OUT(TL_ERROR, ("***** weird entry: %.128s\n", (PCSZ)FilterText(sLine)));
				sPhrase = cSPACE;
				}

			//	Add the rest of the line to the (empty) entry
			CString sRest = sLine.Mid(nEndPhrase);
			sEntry += sRest;

			//	Stop if the phrase range was exceeded
			CHAR cLetter = ToUpper(sPhrase[0]);
			if (cLetter > cLastLetter) {
				DEBUG_OUT(TL_BASIC, ("Last phrase of %c encountered\n", cLastLetter));
				break;
				}
			if (lCount > lMaxEntries) {
				DEBUG_OUT(TL_BASIC, ("Phrase limit %ld reached\n", lMaxEntries));
				break;
				}
			if (cLetter > cCurrentLetter) {
				DEBUG_OUT(TL_BASIC, ("First phrase of %c at offset 0x%08x\n", 
									 cLastLetter, (dict.GetPosition() - sLine.GetLength())));
				bNewLetter = TRUE;
				cCurrentLetter = cLetter;
				}
			}

		//	Otherwise just add the line to the entry
		else {
			sEntry += sLine;
			}
		}

	// clean up shop
	if (pLookup != NULL) {
		pLookup->CleanupSimpleLookup();
		pLookup = NULL;
		}

	dict.Close();
	output.Close();

	return (0);
}
