//	MakeBlockIndex
//
//	This program produces a simple block index (BNX) for the TRL IDX file. The 
//	resulting BNX file gives the last phrase in each chunk of the IDX file. Since 
//	the IDX file actually contains several indices, parameters are required for 
//	specifying the starting and ending offsets for the scan.
//

#include "commonxx.h"
#include "TrlPhrase.h"
#include "TrlIndexInfo.h"
#include "TrlLookup.h"

//	Constants
//

//	Globals
//
CHAR szTrlDataDir[PATH_MAX] = "F:\\DATA";		// source directory for TRL


//	Function prototypes
//
BOOL 	CookIdxFile (PIDXFILE_INFO pxi, PSTR pszFile);
BOOL 	CookBlock (PIDXFILE_INFO pxi, PBYTE pBlock, LONG lOffset, PINDEX pIndex);
BOOL 	WriteIndex (PIDXFILE_INFO pxi, PINDEX pIndex, INT nNumBlocks);


INT main( INT nArgc, PSTR apszArgv[] )
{
    INT nPos;                           // position in the list            
    BOOL bDone;                         // loop termination indicator      
    IDXFILE_INFO xi;                    // Idx file information structure  
    CHAR szFile[PATH_MAX];

	//	Initialize component modules
	HINSTANCE hInstance = NULL;			// TODO: use API to get hInstance?
	hInstance = (HINSTANCE)GetModuleHandle(NULL);
    InitCommon(hInstance, NULL, szLIB_NAME);
    INIT_DEBUGGING_(szLOOKUP_INI);

    // display help information, if no arguments or help indicator
    if ((nArgc == 1) || ((nArgc == 2) && EQUAL_STR(apszArgv[1],"/?"))) {
        fprintf(stderr, "\n" );
        fprintf(stderr, "The comand-line format is MakeBlockIndex <options> <file spec> ...\n");
        fprintf(stderr, "\n" );
        fprintf(stderr, "   <options>:  /o<file name>   file name of block index file (BNX)\n");
        fprintf(stderr, "               /d              run in debugging mode\n");
        fprintf(stderr, "               /a              produce ASCII index (use before /o)\n");
        fprintf(stderr, "               /s              starting offset\n");
        fprintf(stderr, "               /e              ending offset\n");
        fprintf(stderr, "               /@              source directory\n");
#if _DEBUG_MODE
        fprintf(stderr, "               /t<level>     	trace level\n");
#endif
        fprintf(stderr, "\n" );
        fprintf(stderr, "   <file name>:   base name of TRL (Toolworks Reference Library) IDX file\n" );
#if _DEBUG_MODE
        fprintf(stderr, "\n\n" );
        fprintf(stderr, "Debug build: notify TPO of aberations\n");
        fprintf(stderr, "Last compiled: %s\n", __DATE__);
        fprintf(stderr, "Last modified: %s\n", __TIMESTAMP__);
#endif
        fprintf(stderr, "\n\n" );
        exit(1);
        }

#ifdef QUICK_WIN
	_wsetexit(_WINEXITNOPERSIST);
#endif

    // initialize variables
    bDone = FALSE;
    xi.hIdxFile = NULL;
    xi.hIndex = NULL;
    xi.hAsciiIndex = NULL;
    xi.bAsciiIndex = FALSE;
    xi.bDebugMode = FALSE;
    xi.lStartOffset = 0;
    xi.lEndOffset = LONG_MAX;

	//	Get the directory for the TRL files
    INT nLen = GetPrivateProfileString(szSOURCE, szDATA_DIR, GetCurrentDir(), szTrlDataDir, STR_LEN(szTrlDataDir), szLOOKUP_INI);
    AddPathDelim(szTrlDataDir, nLen);

    // process the command-line arguments
    for (nPos=1; nPos<nArgc && !bDone; ++nPos) {

        // process an option switch
        if (apszArgv[nPos][0]=='/') {
        	CHAR cOption = apszArgv[nPos][1];
        	PSTR pszRem = apszArgv[nPos] + 2;
            switch (tolower(cOption)) {
                case 'a':
                    xi.bAsciiIndex = TRUE;
                    xi.bDebugMode = TRUE;
                    break;
                case 'd':
                    xi.bDebugMode = TRUE;
                    break;
                case 'o':
                    sprintf(szFile, "%.8s.BNX", pszRem);
                    xi.hIndex = MyOpenFile(szFile, "w", xi.hIndex, NULL, TRUE);
                    if (xi.bAsciiIndex) {
                        sprintf(szFile, "%.8s.ANX", pszRem);
                        xi.hAsciiIndex = MyOpenFile(szFile, "w", xi.hAsciiIndex, NULL, TRUE);
                        }
                    break;
                case 's':
                	xi.lStartOffset = strtol(pszRem, NULL, 0);
                	Assert(xi.lStartOffset >= 0);
                	break;
                case 'e':
                	xi.lEndOffset = strtol(pszRem, NULL, 0);
                	Assert(xi.lEndOffset >= 0);
                	break;
                case 't': {
					INT nTraceLevel = atoi(pszRem);
					SetTraceLevel(nTraceLevel);
                    break;
					}
                case '@':
                	strcpy(szTrlDataDir, pszRem);
                	break;
                default:
                    fprintf(stderr, "Invalid option: %s\n", apszArgv[nPos]);
                    break;
                }
            }

        // process the index file
        else {
#if TRUE
			if (xi.hIndex == NULL) {
				PCSZ pszFile = "YeIndex";
	            sprintf(szFile, "%.8s.BNX", pszFile);
	            xi.hIndex = MyOpenFile(szFile, "w", xi.hIndex, NULL, TRUE);
	            if (xi.bAsciiIndex) {
	                sprintf(szFile, "%.8s.ANX", pszFile);
	                xi.hAsciiIndex = MyOpenFile(szFile, "w", xi.hAsciiIndex, NULL, TRUE);
	                }
				}
#endif
            CookIdxFile(&xi, apszArgv[nPos]);
            }
        }

    // display elapsed time
    DEBUG_OUT(TL_DETAILED, ("The elapsed time was %gs\n", (double)clock()/CLOCKS_PER_SEC));

    // return with a successful termination code
    return (0);

}


VOID MakePathSpec (PCSZ pszDir, PCSZ pszFile, PSZ pszBuffer, INT nMaxLen)
{
	StringCopy(pszBuffer, pszDir);
	AddPathDelim(pszBuffer, -1);
	StringAppend(pszBuffer, pszFile);
}

//	Analyze the IDX file to produce a block index (BNX), which contains the
//	last phrase in each of 32K chunk of index blocks. The BNX file facilitates
//	a binary search for a phrase.
//
BOOL CookIdxFile (PIDXFILE_INFO pxi, PSTR pszFile)
{
	DEBUG_OUT(TL_VERBOSE, ("CookIdxFile(%x, %s)\n", pxi, pszFile));
    CHAR szFileName[PATH_MAX];
	CHAR szFile[PATH_MAX];

	// open the IDX file
    strcpy(pxi->szSource, pszFile);
	sprintf(szFileName, "%s.IDX", pxi->szSource);
	MakePathSpec(szTrlDataDir, szFileName, szFile, STR_LEN(szFile));
    //pxi->hIdxFile = fopen(szFile,"rb");
	pxi->hIdxFile = MyOpenFile(szFile, "rb", pxi->hIdxFile, NULL, TRUE);
    if (pxi->hIdxFile == NULL) {
        fprintf(stderr, "Unable to open '%s'\n", szFile);
        return (FALSE);
        }

    fprintf(stderr, "Processing %s\n", szFile);

	// open the DAT file
	sprintf(szFileName, "%s.DAT", pxi->szSource);
	MakePathSpec(szTrlDataDir, szFileName, szFile, STR_LEN(szFile));
    //pxi->hDatFile = fopen(szFile,"rb");
	pxi->hDatFile = MyOpenFile(szFile, "rb", pxi->hDatFile, NULL, TRUE);
    if (pxi->hDatFile == NULL) {
        fprintf(stderr, "Unable to open '%s'\n", szFile);
        return (FALSE);
        }
    pxi->lMaxDataOffset = MyGetFileSize(pxi->hDatFile) - 1;

	//	initialize a bunch of stuff
	//
	PBYTE pBlocks = (PBYTE)malloc((WORD)wGROUP_SIZE);
	OldFatalAssert(pBlocks);

	PINDEX pIndex = (PINDEX)malloc(nMAX_GROUPS * sizeof(INDEX));
	OldFatalAssert(pIndex);
	INT nNumGroups = 0;

	LONG lOffset = pxi->lStartOffset;
	LONG lSize = MyGetFileSize(pxi->hIdxFile);
	pxi->lEndOffset = min(pxi->lEndOffset, (LONG)(lSize - 1));
	Assert(pxi->lEndOffset < lSize);
	Assert(pxi->lStartOffset < pxi->lEndOffset);

	pxi->nBytesLeft = 0;
	*(PLONG)pxi->abyResidual = 0;
	CHAR szNullString[] = "";
	pxi->pszLastPhrase = szNullString;

	if (pxi->bAsciiIndex) {
		WriteText(pxi->hAsciiIndex, "Phrase Index:\n\n");
		}

	INT nBlocksLeft = (INT)(((pxi->lEndOffset - lOffset) + (nBLOCK_SIZE - 1)) / nBLOCK_SIZE);
	SetFilePos(pxi->hIdxFile, lOffset, SEEK_SET);
	pxi->lBlockNum = (lOffset / nBLOCK_SIZE) - 1;
	Assert(pxi->lBlockNum >= 0);

	// analyze the IDX blocks and construct a block index
	//
	while (lOffset < pxi->lEndOffset) {

#if _DEBUG_MODE
		//OldFatalAssert(!(feof(pxi->hIdxFile)));
#endif
		INT nToRead = min(nBLOCKS_PER_GROUP, nBlocksLeft);

		INT nRead = ReadData(pBlocks, nBLOCK_SIZE, nToRead, pxi->hIdxFile);
//		Assert((nRead == nToRead) || feof(pxi->hIdxFile));
		nBlocksLeft -= nRead;

#if defined(QUICK_WIN)
		_wyield();						// be a nice QuickWin app
#endif
		for (INT i = 0; i < nRead; i++) {
			CookBlock(pxi, pBlocks + (i * nBLOCK_SIZE), lOffset, pIndex);
			pxi->lBlockNum++;
			lOffset += nBLOCK_SIZE;
			}

		INT nPhraseLen = strlen(pxi->pszLastPhrase);
		OldFatalAssert(nNumGroups < nMAX_GROUPS);
		PSZ pszPhrase = (PSTR)malloc(nPhraseLen + 1);
		OldFatalAssert(pszPhrase);
		strcpy(pszPhrase, pxi->pszLastPhrase);
		pIndex[nNumGroups].pszPhrase = pszPhrase;
		pxi->pszLastPhrase = pszPhrase;

		nNumGroups++;
		}

	// output the block index (BNX)
	//
	WriteIndex(pxi, pIndex, nNumGroups);

	// clean up the mess
	//
	for (INT i = 0; i < nNumGroups; i++) {
		free((PSZ)pIndex[i].pszPhrase);
		}
	free(pIndex);

    MyCloseFile(pxi->hIndex, NULL);
    if (pxi->bAsciiIndex) {
    	MyCloseFile(pxi->hAsciiIndex, NULL);
    	}
    MyCloseFile(pxi->hIdxFile, NULL);
    MyCloseFile(pxi->hDatFile, NULL);
    
    return (TRUE);
}


//	Scan the TRL IDX block to determine the last phrase represented. In addition,
//	an ascii listing of the index is optionaled produced.
//
BOOL CookBlock (PIDXFILE_INFO pxi, PBYTE pBlock, LONG lOffset, PINDEX pIndex)
{
	PBLOCK_HDR pBlockHdr = (PBLOCK_HDR)pBlock;
	BOOL bBlockOK = IsIndexBlock(pxi, pBlockHdr);

	INT nBlockSize = nBLOCK_SIZE;
	if ((lOffset + nBLOCK_SIZE) > pxi->lEndOffset) {
		nBlockSize = (INT)(pxi->lEndOffset - lOffset);
		Assert(nBlockSize > 0);
		}
	PBYTE pBlockEnd = pBlock + nBlockSize;

	PCSZ pszFirstPhrase = szNA;
	PCSZ pszLastPhrase = pxi->pszLastPhrase;

	if (pxi->bAsciiIndex) {
		WriteTextFmt(pxi->hAsciiIndex, "\nOffset %08lx\n", lOffset);
#if defined(QUICK_WIN)
		_wyield();						// be a nice QuickWin app
#endif
		}

	BOOL bFirstEntry = TRUE;
	
	ENTRY_INFO Entry;
	for (PBYTE pStart = pBlock + sizeof(BLOCK_HDR); pStart < pBlockEnd; pStart = Entry.pNext) {
		BOOL bEntryOK = InterpretEntry(pStart, &Entry, pxi);
		if (!bEntryOK) {
			break;
			}
		if (Entry.nPhraseLen == -1) {
			DEBUG_OUT(TL_BASIC, ("Ignoring bogus entry at offset %lx\n", lOffset));
			continue;
			}

		if (bFirstEntry) {
			pszFirstPhrase = Entry.pszPhrase;
			bFirstEntry = FALSE;
			}

		//	Make note of phrase; exit if reference listing is not needed
       	pszLastPhrase = Entry.pszPhrase;
        if (!pxi->bAsciiIndex || !bBlockOK) {
        	continue;
        	}
		
		//	Determine the number of references and the starting offset
		InterpretEntryRefs(pStart, &Entry, pxi);

		if (pxi->bAsciiIndex) {
			WriteTextFmt(pxi->hAsciiIndex, "%-20s", Entry.pszPhrase);
#if _DEBUG_MODE
			WriteTextFmt(pxi->hAsciiIndex, "<%x %x %x %x %x>\t",
					Entry.pHdr->byRefType, Entry.pHdr->byEntryPos, Entry.pHdr->byCode2,
					Entry.pHdr->byEntryType, Entry.pHdr->byCode3);
#endif
			if (Entry.bContinuation) {
				WriteTextFmt(pxi->hAsciiIndex, "<r>%08lx ", *(PLONG)pxi->abyResidual);
				*(PLONG)pxi->abyResidual = 0;
				}
			for (INT i = 0; i < Entry.nNumRefs; i++) {
				Assert (Entry.plRefs[i] <= pxi->lMaxDataOffset);
				WriteTextFmt(pxi->hAsciiIndex, "%08lx ", Entry.plRefs[i]);
				}
			if (pxi->nBytesLeft > 0) {
				WriteTextFmt(pxi->hAsciiIndex, "<TBC; -%d>", sizeof(LONG) - pxi->nBytesLeft);
				}
			WriteText(pxi->hAsciiIndex, "\n");
			}
		}

	OldFatalAssert((pszFirstPhrase) && (pszLastPhrase));
	DEBUG_OUT(TL_DETAILED, ("%08lx %-20.20s %-20.20s %ld %ld %ld %2x %2x %2x ",
			 lOffset, pszFirstPhrase, pszLastPhrase,
			 pBlockHdr->lCode1, pBlockHdr->lNextBlock, pBlockHdr->lPrevBlock,
			 pBlockHdr->wCode1, pBlockHdr->wCode2, pBlockHdr->wCode3));
#if _DEBUG_MODE
	static CHAR szPreviousLast[256] = " ";
	if (!bBlockOK) {
		DEBUG_OUT(TL_DETAILED, ("<I>"));
		}
	else {
		if (PhraseGT(szPreviousLast, pszFirstPhrase, TRUE)) {
			DEBUG_OUT(TL_DETAILED, ("<S>"));
			}
		strncpy(szPreviousLast, pszLastPhrase, sizeof(szPreviousLast) - 1);
		szPreviousLast[sizeof(szPreviousLast)-1] = NULL_CHAR;
		}
#endif
	DEBUG_OUT(TL_DETAILED, ("\n"));

	if (bBlockOK) {
		pxi->pszLastPhrase = pszLastPhrase;
		}

	return (bBlockOK);
}


//	Write out the block index
//
//		TRL file base (eg, DICT)
//		starting offset in IDX (eg, 0x68800)
//		list of last phrases in each 32K block
//
BOOL WriteIndex (PIDXFILE_INFO pxi, PINDEX pIndex, INT nNumBlocks)
{
	DEBUG_OUT(TL_DETAILED, ("\n\nIndex:\n"));

	WriteTextFmt(pxi->hIndex, "%s\n", pxi->szSource);
	WriteTextFmt(pxi->hIndex, "0x%lx\n", pxi->lStartOffset);
	for (INT i = 0; i < nNumBlocks; i++) {
		WriteTextFmt(pxi->hIndex, "%s\n", pIndex[i].pszPhrase);

		LONG lOffset = pxi->lStartOffset + ((LONG)(i + 1) * wGROUP_SIZE);
		DEBUG_OUT(TL_DETAILED, ("%08lx %s\n", lOffset, pIndex[i].pszPhrase));
		}

	return (TRUE);
}
