/*

  wn.h - header file for use with WordNet searching, morphology 
         and binary search functions

     $Id: wn.h,v 1.40 1995/02/07 16:16:13 wn Exp $
*/

#ifndef _WN_
#define _WN_

/* Data structures used by search code functions. */

/* Structure for index file entry */
typedef struct {
    char *wd;           /* word string */
    char *pos;          /* part of speech */
    int sense_cnt;      /* sense (collins) count */
    int off_cnt;        /* number of offsets */
    long *offset;       /* offsets of synsets containing word */
    int ptruse_cnt;     /* number of pointers used */
    int *ptruse;        /* pointers used */
} Index, *IndexPtr;

/* Structure for data file synset */
typedef struct ss {
    long hereiam;       /* current file position */
    int sstype;         /* type of ADJ synset */
    int fnum;           /* file number that synset comes from */
    char *pos;          /* part of speech */
    int wcount;         /* number of words in synset */
    char **words;       /* words in synset */
    int *wnsense;       /* sense number in wordnet */
    int whichword;      /* which word in synset we're looking for */
    int ptrcount;       /* number of pointers */
    int *ptrtyp;        /* pointer types */
    long *ptroff;       /* pointer offsets */
    int *ppos;          /* pointer part of speech */
    int *pto;           /* pointer 'to' fields */
    int *pfrm;          /* pointer 'from' fields */
    int fcount;         /* number of verb frames */
    int *frmid;         /* frame numbers */
    int *frmto;         /* frame 'to' fields */
    char *defn;         /* synset gloss (definition) */

    /* these fields are used if a data structure is returned
       instead of a text buffer */

    struct ss *nextss;      /* ptr to next synset containing searchword */
    int searchtype;     /* type of search performed */
    struct ss *ptrlist;     /* ptr to synset list result of search */
    char *headword;     /* if pos is "s", this is cluster head word */
    short headsense;        /* sense number of headword */
} Synset, *SynsetPtr;


/*** Search and database functions (search.c) ***/

/* Primry search algorithm for use with user interfaces */
extern char *findtheinfo(char *, int, int, int);    

/* Primary search algorithm for use with programs (returns data structure) */
extern SynsetPtr findtheinfo_ds(char *, int, int, int); 

/* Set bit for each search type that is valid for the search word
   passed and return bit mask. */
extern unsigned long is_defined(char *, int); 

/* Set bit for each POS that search word is in.  0 returned if
   word is not in WordNet. */
extern unsigned int in_wn(char *, int); 

/* Find word in index file and return parsed entry in data structure.
   Input word must be exact match of string in database. */
extern IndexPtr index_lookup(char *, int); 

/* 'smart' search of index file.  Find word in index file, trying different
   techniques - replace hyphens with underscores, replace underscores with
   hyphens, strip hyphens and underscores, strip periods. */
extern IndexPtr getindex(char *, int);  

/* Read synset from data file at byte offset passed and return parsed
   entry in data structure. */
extern SynsetPtr read_synset(int, long, char *);

/* Read synset at current byte offset in file and return parsed entry
   in data structure. */
extern SynsetPtr parse_synset(FILE *, int, char *); 

/* Free a synset linked list allocated by findtheinfo_ds() */
extern void free_syns(SynsetPtr);   

/* Free a synset */
extern void free_synset(SynsetPtr); 

/* Free an index structure */
extern void free_index(IndexPtr);   

/* Recursive search algorithm to trace a pointer tree and return results
   in linked list of data structures. */
SynsetPtr traceptrs_ds(SynsetPtr, int, int, int);

/* Do requested search on synset passed, returning output in buffer. */
extern char *do_trace(SynsetPtr, int, int, int);

/* Trace cousin tree */
extern int trace_cousin(SynsetPtr, unsigned long, int);

/* Read cousin.tops file into array */
extern void read_cousintops();  

/*** Morphology functions (morph.c) ***/

/* Open exception list files */
extern int morphinit(); 

#if 1
//	Close exception list files
int morphclose (void);
#endif

/* Close exception list files and reopen */
extern int re_morphinit();  

/* Try to find baseform (lemma) of word or collocation in POS. */
extern char *morphstr(char *, int); 

/* Try to find baseform (lemma) of individual word in POS. */
extern char *morphword(char *, int);    

/*** Utility functions (wnutil.c) ***/

/* Top level function to open database files, initialize wn_filenames,
   and open exeception lists. */
extern int wninit();        

#if 1
extern int wnclose (void);
#endif

/* Top level function to close and reopen database files, initialize
   wn_filenames and open exception lists. */
extern int re_wninit(); 

/* Count the number of underscore or space separated words in a string. */
extern int cntwords(char *, char);      

/* Convert string to lower case remove trailing adjective marker if found */
extern char *strtolower(char *);    

/* Convert string passed to lower case */
extern char *ToLowerCase(char *);   

/* Replace all occurrences of 'from' with 'to' in 'str' */
extern char *strsubst(char *, char, char);  

/* Return pointer code for pointer type characer passed. */
extern int getptrtype(char *);  

/* Return part of speech code for string passed */
extern int getpos(char *);      

extern int GetPartOfSpeach (char *s);

/* Return synset type code for string passed. */
extern int getsstype(char *);       

/* Reconstruct synset from synset pointer and return ptr to buffer */
extern char *FmtSynset(SynsetPtr, int); 

/* Find string for 'searchstr' as it is in index file */
extern char *GetWNStr(char *, int);

/* Pass in string for POS, return corresponding integer value */
extern int StrToPos(char *);

/* Return synset for sense key passed. */
extern SynsetPtr GetSynsetForSense(char *);

/* Find offset of sense key in data file */
extern long GetDataOffset(char *);

/* Find polysemy (collins) count for sense key passed. */
extern int GetPolyCount(char *);

/* Return word part of sense key */
extern char *GetWORD(char *);

/* Return POS code for sense key passed. */
extern int GetPOS(char *);

/* Convert WordNet sense number passed of IndexPtr entry to sense key. */
extern char *WNSnsToStr(IndexPtr, int);

/* Search for string and/or baseform of word in database and return
   index structure for word if found in database. */
extern IndexPtr GetValidIndexPointer(char *, int);

#ifdef NEXT
extern char *strdup();      /* duplicates a string (needed for NeXT) */
#endif

/* Return sense number in database for word and lexsn passed. */
int GetWNSense(char *word, char *lexsn);

/*** Binary search functions (binsearch.c) ***/

/* General purpose binary search function to search for key as first
   item on line in open file.  Item is delimited by space. */
extern char *bin_search(char *, FILE *);

/* Copy contents from one file to another. */
extern void copyfile(FILE *, FILE *);

/* Function to replace a line in a file.  Returns the original line,
   or NULL in case of error. */
extern char *replace_line(char *, char *, FILE *);

/* Find location to insert line at in file.  If line with this
   key is already in file, return NULL. */
extern char *insert_line(char *, char *, FILE *);

#endif /*_WN_*/

/*
  Revision log:

  $Log: wn.h,v $
 * Revision 1.40  1995/02/07  16:16:13  wn
 * removed do_traceds
 *
 * Revision 1.39  1995/02/07  16:11:31  wn
 * added prototypes, cleanup
 *
 * Revision 1.38  1995/01/17  20:56:29  wn
 * added prototypes for lib functions
 *
 * Revision 1.37  1994/11/30  16:50:31  wn
 * added read_cousintops and trace_cousins
 *
 * Revision 1.36  1994/10/07  14:39:50  wn
 * added parse_synset
 *
 * Revision 1.35  1994/08/03  17:24:47  wn
 * removed StrToSense and SenseToStr
 *
 * Revision 1.34  1993/12/22  14:23:31  wn
 * added WNESCORT
 *
 * Revision 1.32  93/07/30  13:09:57  wn
 * added strdup for NeXT
 * 
 * Revision 1.31  93/07/29  13:45:59  wn
 * *** empty log message ***
 * 
 * Revision 1.30  93/07/29  13:42:02  wn
 * *** empty log message ***
 * 
 * Revision 1.29  93/07/28  16:32:52  wn
 * added re_*init
 * 
 * Revision 1.28  93/06/09  14:52:58  wn
 * added in_wn
 * 
 * Revision 1.27  93/05/05  13:26:50  wn
 * 
 * 
 * Revision 1.26  93/05/04  13:39:41  wn
 * added insert_line, replace_likne
 * 
 * Revision 1.25  93/04/13  13:48:22  wn
 * added free_synset and free_index
 * 
 * 
 * Revision 1.24  93/03/11  16:35:15  wn
 * added #ifndef to protect from multiple inclusions
 * 
 * Revision 1.23  93/01/26  13:56:14  wn
 * added extern for getindex
 * 
 * Revision 1.22  92/09/09  13:26:45  wn
 * removed search code globals variables - moved to wnglobals.h

*/
